"""
Collection of functions to administer Monte Carlo simulation.
"""

import os as os
import numpy as np
import pandas as pd
import multiprocessing as mp
from scipy.stats import uniform, triang, truncnorm

from pyscnomics.tools.summary import get_summary
from pyscnomics.io.aggregator import Aggregate
from pyscnomics.optimize.adjuster import AdjustData

from pyscnomics.contracts.project import BaseProject
from pyscnomics.contracts.costrecovery import CostRecovery
from pyscnomics.contracts.grossplit import GrossSplit
from pyscnomics.contracts.transition import Transition


class MonteCarloException(Exception):
    """ Exception to be raised for a misuse of MonteCarlo class """

    pass


def get_montecarlo_data(data: Aggregate) -> dict:
    """
    Prepares Monte Carlo simulation data from the given Aggregate object.

    Parameters
    ----------
    data: Aggregate
        The input data containing Monte Carlo simulation data.

    Returns
    -------
    dict
        A dictionary containing the prepared Monte Carlo simulation data.

    Raises
    ------
    MonteCarloException
        If any of the required data is missing or inconsistent.
    """
    # Prepare a container to store all the necessary data
    uncertainty_data = {}

    # Prepare attribute run_number
    if pd.isna(data.montecarlo_data.run_number):
        raise MonteCarloException(
            f"The data for run number is missing. Please check the input data."
        )

    uncertainty_data["run_number"] = int(data.montecarlo_data.run_number)

    # Prepare attribute parameter
    if not isinstance(data.montecarlo_data.parameter, list):
        raise MonteCarloException(
            f"Attribute parameter must be provided in a list datatype, "
            f"not {data.montecarlo_data.parameter.__class__.__qualname__}"
        )

    uncertainty_data["parameter"] = data.montecarlo_data.parameter

    # Prepare attribute distribution
    if not isinstance(data.montecarlo_data.distribution, list):
        raise MonteCarloException(
            f"Attribute distribution must be provided in a list datatype, "
            f"not {data.montecarlo_data.distribution.__class__.__qualname__}"
        )

    uncertainty_data["distribution"] = data.montecarlo_data.distribution

    # Prepare attribute min_values
    if not isinstance(data.montecarlo_data.min_values, np.ndarray):
        raise MonteCarloException(
            f"Attribute min_values must be provided in a numpy ndarray datatype, "
            f"not {data.montecarlo_data.min_values.__class__.__qualname__}"
        )

    min_values_nan = list(filter(lambda i: pd.isna(i), data.montecarlo_data.min_values))
    if len(min_values_nan) > 0:
        raise MonteCarloException(
            f"Minimum values data are incomplete. Please check the input data."
        )

    uncertainty_data["min_values"] = data.montecarlo_data.min_values.astype(np.float_)

    # Prepare attribute mean_values
    if not isinstance(data.montecarlo_data.mean_values, np.ndarray):
        raise MonteCarloException(
            f"Attribute mean_values must be provided in a numpy ndarray datatype, "
            f"not {data.montecarlo_data.mean_values.__class__.__qualname__}"
        )

    mean_values_nan = list(filter(lambda i: pd.isna(i), data.montecarlo_data.mean_values))
    if len(mean_values_nan) > 0:
        raise MonteCarloException(
            f"Mean values data are incomplete. Please check the input data."
        )

    uncertainty_data["mean_values"] = data.montecarlo_data.mean_values.astype(np.float_)

    # Prepare attribute max_values
    if not isinstance(data.montecarlo_data.max_values, np.ndarray):
        raise MonteCarloException(
            f"Attribute max_values must be provided in a numpy ndarray datatype, "
            f"not {data.montecarlo_data.max_values.__class__.__qualname__}"
        )

    max_values_nan = list(filter(lambda i: pd.isna(i), data.montecarlo_data.max_values))
    if len(max_values_nan) > 0:
        raise MonteCarloException(
            f"Maximum values data are incomplete. Please check the input data."
        )

    uncertainty_data["max_values"] = data.montecarlo_data.max_values.astype(np.float_)

    # Prepare attribute standard deviation
    if not isinstance(data.montecarlo_data.std_dev, np.ndarray):
        raise MonteCarloException(
            f"Attribute std_dev must be provided in a numpy ndarray datatype, "
            f"not {data.montecarlo_data.std_dev.__class__.__qualname__}"
        )

    for i, val in enumerate(uncertainty_data["distribution"]):
        if val == "Uniform" or val == "Triangular":
            data.montecarlo_data.std_dev[i] = 0.0

    std_dev_nan = list(filter(lambda i: pd.isna(i), data.montecarlo_data.std_dev))
    if len(std_dev_nan) > 0:
        raise MonteCarloException(
            f"Std deviation data are incomplete. Please check the input data."
        )

    uncertainty_data["std_dev"] = data.montecarlo_data.std_dev.astype(np.float_)

    # Throw exception for inappropriate conditions
    for i, val in enumerate(uncertainty_data["min_values"]):
        if (
            uncertainty_data["min_values"][i] > uncertainty_data["max_values"][i]
            or uncertainty_data["min_values"][i] > uncertainty_data["mean_values"][i]
            or uncertainty_data["mean_values"][i] > uncertainty_data["max_values"][i]
        ):
            raise MonteCarloException(
                f"Inappropriate minimum, mean, or maximum value data. "
                f"Please check the input data."
            )

    # Identify zero mean values
    mean_values_not_zero = np.argwhere(uncertainty_data["mean_values"] != 0.0).ravel()

    return uncertainty_data, mean_values_not_zero


def get_multipliers_montecarlo(
    run_number: int,
    distribution: str,
    min_value: float,
    mean_value: float,
    max_value: float,
    std_dev: float,
) -> np.ndarray:
    """
    Generate an array of multipliers for Monte Carlo simulation based on the specified distribution.

    Parameters
    ----------
    run_number: int
        Number of runs.
    distribution: str
        Type of distribution ("Uniform", "Triangular", or "Normal").
    min_value: float
        Minimum value for the distribution.
    mean_value: float
        Mean (or central value) for the distribution.
    max_value: float
        Maximum value for the distribution.
    std_dev: float
        Standard deviation for the normal distribution.

    Returns
    -------
    multipliers: np.ndarray
        Array of multipliers generated using Monte Carlo simulation.

    Notes
    -----
    - For "Uniform" distribution, the function uses a uniform random variable.
    - For "Triangular" distribution, the function uses a triangular random variable.
    - For "Normal" distribution, the function uses a truncated normal random variable.
    """
    # Uniform distribution
    if distribution == "Uniform":
        # Modify minimum and maximum values
        params = [min_value, max_value]
        attrs_updated = {
            key: float(params[i] / mean_value)
            for i, key in enumerate(["min_value", "max_value"])
        }

        # Determine multipliers
        multipliers = uniform.rvs(
            loc=attrs_updated["min_value"],
            scale=attrs_updated["max_value"] - attrs_updated["min_value"],
            size=run_number,
        )

    # Triangular distribution
    elif distribution == "Triangular":
        # Modify minimum and maximum values
        params = [min_value, mean_value, max_value]
        attrs_updated = {
            key: float(params[i] / mean_value)
            for i, key in enumerate(["min_value", "mean_value", "max_value"])
        }

        # Determine mode (central point)
        c = (
            (attrs_updated["mean_value"] - attrs_updated["min_value"]) /
            (attrs_updated["max_value"] - attrs_updated["min_value"])
        )

        # Determine multipliers
        multipliers = triang.rvs(
            c=c,
            loc=attrs_updated["min_value"],
            scale=attrs_updated["max_value"] - attrs_updated["min_value"],
            size=run_number,
        )

    # Normal distribution
    elif distribution == "Normal":
        # Modify minimum and maximum values
        params = [min_value, mean_value, max_value, std_dev]
        attrs_updated = {
            key: float(params[i] / mean_value)
            for i, key in enumerate(["min_value", "mean_value", "max_value", "std_dev"])
        }

        # Determine z-values
        zvalues = {
            key: float(
                (attrs_updated[key] - attrs_updated["mean_value"]) / attrs_updated["std_dev"]
            )
            for key in ["min_value", "mean_value", "max_value"]
        }

        # Determine multipliers
        multipliers_init = truncnorm.rvs(
            a=zvalues["min_value"],
            b=zvalues["max_value"],
            loc=zvalues["mean_value"],
            scale=1,
            size=run_number,
        )

        multipliers = (multipliers_init * attrs_updated["std_dev"]) + attrs_updated["mean_value"]

    else:
        raise MonteCarloException(
            f"The type of distribution is unavailable. "
            f"Please select type of distribution between: "
            f"(i) Uniform, (ii) Triangular, or (iii) Normal."
        )

    return multipliers


def run_montecarlo(
    contract: BaseProject | CostRecovery | GrossSplit | Transition,
    contract_arguments: dict,
    summary_arguments: dict,
) -> tuple:
    """
    Function to run Monte Carlo simulation.

    Parameters
    ----------
    contract: BaseProject | CostRecovery | GrossSplit | Transition
        The contract type for the project.
    contract_arguments: dict
        Arguments required for running the specified contract.
    summary_arguments: dict
        Arguments for generating the project summary.

    Returns
    -------
    tuple:
        A tuple containing the following project metrics:
        - NPV (Net Present Value)
        - IRR (Internal Rate of Return)
        - PI (Profitability Index)
        - POT (Pay Out Time)
        - Gov Take (Government Take)
        - Net Share (Contractor's Net Share)
    """
    contract.run(**contract_arguments)
    contract_summary = get_summary(**summary_arguments)

    return (
        contract_summary["ctr_npv"],
        contract_summary["ctr_irr"],
        contract_summary["ctr_pi"],
        contract_summary["ctr_pot"],
        contract_summary["gov_take"],
        contract_summary["ctr_net_share"],
    )


def execute_montecarlo(
    data: Aggregate,
    workbook_path: str,
    uncertainty_data: dict,
    target: list,
    mult: np.ndarray,
    threshold_parallel: int = 500,
) -> np.ndarray:
    """
    A target function to run MonteCarlo simulation.

    Parameters
    ----------
    data: Aggregate
        The aggregate data for the simulation.
    workbook_path: str
        The path to the workbook containing data.
    uncertainty_data: dict
        Data associated with uncertainty analysis.
    target: list
        A list of targets for the simulation.
    mult: np.ndarray
        A numpy array containing multipliers.
    threshold_parallel: int
        A threshold value to activate parallel computation.

    Returns
    -------
    results: np.ndarray
        Results of the Monte Carlo simulation as a numpy array.

    Notes
    -----
    The engine executes parallel computing for number of simulation run larger than 1_000.
    Instead, serial computing is adopted.
    """
    # Generate project instances to be run in the simulation
    args = ["psc", "psc_arguments", "summary_arguments"]
    params = {arg: [0] * uncertainty_data["run_number"] for arg in args}

    for n in range(uncertainty_data["run_number"]):
        (
            params["psc"][n],
            params["psc_arguments"][n],
            params["summary_arguments"][n],
        ) = AdjustData(
            data=data,
            workbook_path=workbook_path,
            multipliers=mult[n, :],
        ).activate()

    # Execute MonteCarlo simulation
    results = np.zeros([uncertainty_data["run_number"], len(target)], dtype=np.float_)

    # Run serial for number of simulation less than 1000
    if uncertainty_data["run_number"] <= threshold_parallel:
        for n in range(uncertainty_data["run_number"]):
            results[n, :] = run_montecarlo(
                contract=params["psc"][n],
                contract_arguments=params["psc_arguments"][n],
                summary_arguments=params["summary_arguments"][n],
            )

    # Run parallel for number of simulation larger than 1000
    else:
        combined_iters = []
        for n in range(uncertainty_data["run_number"]):
            combined_iters.append(
                (
                    params["psc"][n],
                    params["psc_arguments"][n],
                    params["summary_arguments"][n],
                )
            )

        pl = mp.Pool(processes=int(os.cpu_count() - 1))
        results_init = pl.starmap(run_montecarlo, combined_iters)
        pl.close()
        pl.join()

        # Arrange the results into the appointed data structure
        for n in range(uncertainty_data["run_number"]):
            results[n, :] = results_init[n]

    return results



################################################ Uncertainty Detached ################################################
"""
This Code is made to detached the uncertainty module in the previous version which depended with the excel 
into fully able to be run in python. 
"""

from pyscnomics.econ.selection import UncertaintyDistribution
from pyscnomics.optimize.sensitivity import _adjust_element_single_contract

def get_multipliers_monte(
    run_number: int,
    distribution: UncertaintyDistribution,
    min_value: float,
    mean_value: float,
    max_value: float,
    std_dev: float,
) -> np.ndarray:
    """
    Generate an array of multipliers for Monte Carlo simulation based on the specified distribution.

    Parameters
    ----------
    run_number: int
        Number of runs.
    distribution: str
        Type of distribution ("Uniform", "Triangular", or "Normal").
    min_value: float
        Minimum value for the distribution.
    mean_value: float
        Mean (or central value) for the distribution.
    max_value: float
        Maximum value for the distribution.
    std_dev: float
        Standard deviation for the normal distribution.

    Returns
    -------
    multipliers: np.ndarray
        Array of multipliers generated using Monte Carlo simulation.

    Notes
    -----
    - For "Uniform" distribution, the function uses a uniform random variable.
    - For "Triangular" distribution, the function uses a triangular random variable.
    - For "Normal" distribution, the function uses a truncated normal random variable.
    """
    # Uniform distribution
    if distribution == UncertaintyDistribution.UNIFORM:
        # Modify minimum and maximum values
        params = [min_value, max_value]
        attrs_updated = {
            key: float(params[i] / mean_value)
            for i, key in enumerate(["min_value", "max_value"])
        }

        # Determine multipliers
        multipliers = uniform.rvs(
            loc=attrs_updated["min_value"],
            scale=attrs_updated["max_value"] - attrs_updated["min_value"],
            size=run_number,
        )

    # Triangular distribution
    elif distribution == UncertaintyDistribution.TRIANGULAR:
        # Modify minimum and maximum values
        params = [min_value, mean_value, max_value]
        attrs_updated = {
            key: float(params[i] / mean_value)
            for i, key in enumerate(["min_value", "mean_value", "max_value"])
        }

        # Determine mode (central point)
        c = (
            (attrs_updated["mean_value"] - attrs_updated["min_value"]) /
            (attrs_updated["max_value"] - attrs_updated["min_value"])
        )

        # Determine multipliers
        multipliers = triang.rvs(
            c=c,
            loc=attrs_updated["min_value"],
            scale=attrs_updated["max_value"] - attrs_updated["min_value"],
            size=run_number,
        )

    # Normal distribution
    elif distribution == UncertaintyDistribution.NORMAL:
        # Modify minimum and maximum values
        params = [min_value, mean_value, max_value, std_dev]
        attrs_updated = {
            key: float(params[i] / mean_value)
            for i, key in enumerate(["min_value", "mean_value", "max_value", "std_dev"])
        }

        # Determine z-values
        zvalues = {
            key: float(
                (attrs_updated[key] - attrs_updated["mean_value"]) / attrs_updated["std_dev"]
            )
            for key in ["min_value", "mean_value", "max_value"]
        }

        # Determine multipliers
        multipliers_init = truncnorm.rvs(
            a=zvalues["min_value"],
            b=zvalues["max_value"],
            loc=zvalues["mean_value"],
            scale=1,
            size=run_number,
        )

        multipliers = (multipliers_init * attrs_updated["std_dev"]) + attrs_updated["mean_value"]

    else:
        raise MonteCarloException(
            f"The type of distribution is unavailable. "
            f"Please select type of distribution between: "
            f"(i) Uniform, (ii) Triangular, or (iii) Normal."
        )

    return multipliers

def min_mean_max_retriever(
        contract: BaseProject | CostRecovery | GrossSplit | Transition,
):
    # Retrieve Min, Mean, and Max for CAPEX
    min_capex = np.min(np.concatenate(
        (contract.capital_cost_total.cost, contract.intangible_cost_total.cost)))
    mean_capex = np.average(np.concatenate(
        (contract.capital_cost_total.cost, contract.intangible_cost_total.cost)))
    max_capex = np.max(np.concatenate(
        (contract.capital_cost_total.cost, contract.intangible_cost_total.cost)))

    # Retrieve Min, Mean, and Max for OPEX
    min_opex = np.min(np.concatenate(
        (contract.opex_total.cost, contract.asr_cost_total.cost,
         contract.lbt_cost_total.cost, contract.cost_of_sales_total.cost)))
    mean_opex = np.average(np.concatenate(
        (contract.opex_total.cost, contract.asr_cost_total.cost,
         contract.lbt_cost_total.cost, contract.cost_of_sales_total.cost)))
    max_opex = np.max(np.concatenate(
        (contract.opex_total.cost, contract.asr_cost_total.cost,
         contract.lbt_cost_total.cost, contract.cost_of_sales_total.cost)))

    # Retrieve Min, Mean, and Max for Oil Price
    min_oil_price = np.min(contract._oil_lifting.price)
    mean_oil_price = np.average(contract._oil_lifting.price)
    max_oil_price = np.max(contract._oil_lifting.price)

    # Retrieve Min, Mean, Max for Gas Price
    min_gas_price = np.min(contract._gas_lifting.price)
    mean_gas_price = np.average(contract._gas_lifting.price)
    max_gas_price = np.max(contract._gas_lifting.price)

    # Retrieve Min, Mean, Max for Oil Lifting
    min_oil_lifting = np.min(contract._oil_lifting.lifting_rate)
    mean_oil_lifting = np.average(contract._oil_lifting.lifting_rate)
    max_oil_lifting = np.max(contract._oil_lifting.lifting_rate)

    # Retrieve Min, Mean, Max for Gas Lifting
    min_gas_lifting = np.min(contract._gas_lifting.lifting_rate)
    mean_gas_lifting = np.average(contract._gas_lifting.lifting_rate)
    max_gas_lifting = np.max(contract._gas_lifting.lifting_rate)

    return {
        'min_capex': min_capex,
        'mean_capex': mean_capex,
        'max_capex': max_capex,
        'min_opex': min_opex,
        'mean_opex': mean_opex,
        'max_opex': max_opex,
        'min_oil_price': min_oil_price,
        'mean_oil_price': mean_oil_price,
        'max_oil_price': max_oil_price,
        'min_gas_price': min_gas_price,
        'mean_gas_price': mean_gas_price,
        'max_gas_price': max_gas_price,
        'min_oil_lifting': min_oil_lifting,
        'mean_oil_lifting': mean_oil_lifting,
        'max_oil_lifting': max_oil_lifting,
        'min_gas_lifting': min_gas_lifting,
        'mean_gas_lifting': mean_gas_lifting,
        'max_gas_lifting': max_gas_lifting,
    }


def uncertainty_psc(
        contract: BaseProject | CostRecovery | GrossSplit | Transition,
        contract_arguments: dict,
        summary_arguments: dict,
        run_number: int = 100,
        min_capex: float = None,
        mean_capex: float = None,
        max_capex: float = None,
        min_opex: float = None,
        mean_opex: float = None,
        max_opex: float = None,
        min_oil_price: float = None,
        mean_oil_price: float = None,
        max_oil_price: float = None,
        min_gas_price: float = None,
        mean_gas_price: float = None,
        max_gas_price: float = None,
        min_oil_lifting: float = None,
        mean_oil_lifting: float = None,
        max_oil_lifting: float = None,
        min_gas_lifting: float = None,
        mean_gas_lifting: float = None,
        max_gas_lifting: float = None,
        capex_stddev: float = 1.25,
        opex_stddev: float = 1.25,
        oil_price_stddev: float = 1.25,
        gas_price_stddev: float = 1.25,
        oil_lifting_stddev: float = 1.25,
        gas_lifting_stddev: float = 1.25,
        capex_distribution: UncertaintyDistribution = UncertaintyDistribution.NORMAL,
        opex_distribution: UncertaintyDistribution = UncertaintyDistribution.NORMAL,
        oil_price_distribution: UncertaintyDistribution = UncertaintyDistribution.NORMAL,
        gas_price_distribution: UncertaintyDistribution = UncertaintyDistribution.NORMAL,
        oil_lifting_distribution: UncertaintyDistribution = UncertaintyDistribution.NORMAL,
        gas_lifting_distribution: UncertaintyDistribution = UncertaintyDistribution.NORMAL,
        dataframe_output: bool = True,
) -> tuple:
    """
    The function to get the uncertainty analysis (Monte Carlo) of a contract.

    Parameters
    ----------
    contract: BaseProject | CostRecovery | GrossSplit | Transition
        The contract that the uncertainty will be retrieved.
    contract_arguments: dict
        The contract arguments of the contract.
    summary_arguments: dict
        The summary arguments.
    run_number: int = 100
        The total run number for the monte carlo.
    min_capex: float = None
        Minimum value for the capital cost component.
    mean_capex: float = None
        Average value for the capital cost component.
    max_capex: float = None
        Maximum value for the capital cost component.
    min_opex: float = None
        Minimum value for the opex cost component.
    mean_opex: float = None
        Average value for the opex cost component.
    max_opex: float = None
        Maximum value for the opex cost component.
    min_oil_price: float = None
        Minimum value for the price of oil lifting.
    mean_oil_price: float = None
        Average value for the price of oil lifting.
    max_oil_price: float = None
        Maximum value for the price of oil lifting.
    min_gas_price: float = None
        Minimum value for the price of gas lifting.
    mean_gas_price: float = None
        Average value for the price of gas lifting.
    max_gas_price: float = None
        Maximum value for the price of gas lifting.
    min_oil_lifting: float = None
        Minimum value for the oil lifting.
    mean_oil_lifting: float = None
        Average value for the oil lifting.
    max_oil_lifting: float = None
        Maximum value for the oil lifting.
    min_gas_lifting: float = None
        Minimum value for the gas lifting.
    mean_gas_lifting: float = None
        Average value for the gas lifting.
    max_gas_lifting: float = None
        Maximum value for the gas lifting.
    capex_stddev: float = 1.25
        The standard deviation of the capital cost element which will be used in monte carlo
    opex_stddev: float = 1.25
        The standard deviation of the opex cost element which will be used in monte carlo
    oil_price_stddev: float = 1.25
        The standard deviation of the oil price element which will be used in monte carlo
    gas_price_stddev: float = 1.25
        The standard deviation of the gas price element which will be used in monte carlo
    oil_lifting_stddev: float = 1.25
        The standard deviation of the oil lifting element which will be used in monte carlo
    gas_lifting_stddev: float = 1.25
        The standard deviation of the gas lifting element which will be used in monte carlo
    capex_distribution: UncertaintyDistribution = UncertaintyDistribution.NORMAL
        The distribution option which will be used in generating capex values.
    opex_distribution: UncertaintyDistribution = UncertaintyDistribution.NORMAL
        The distribution option which will be used in generating opex values.
    oil_price_distribution: UncertaintyDistribution = UncertaintyDistribution.NORMAL
        The distribution option which will be used in generating oil price values.
    gas_price_distribution: UncertaintyDistribution = UncertaintyDistribution.NORMAL
        The distribution option which will be used in generating gas price values.
    oil_lifting_distribution: UncertaintyDistribution = UncertaintyDistribution.NORMAL
        The distribution option which will be used in generating oil lifting values.
    gas_lifting_distribution: UncertaintyDistribution = UncertaintyDistribution.NORMAL
        The distribution option which will be used in generating gas lifting values.
    dataframe_output: bool = True
        The option whether the output in a dataframe form or dictionary.
    Returns
    -------
    out: tuple
        Uncertainty result,
        Group of the uncertainty result [P10, P50 AND P90]
    """
    # Retrieve the Min Max and Average Values
    min_max_mean_original = min_mean_max_retriever(contract=contract)

    min_max_mean_std = {
        'min_capex': min_capex,
        'mean_capex': mean_capex,
        'max_capex': max_capex,
        'min_opex': min_opex,
        'mean_opex': mean_opex,
        'max_opex': max_opex,
        'min_oil_price': min_oil_price,
        'mean_oil_price': mean_oil_price,
        'max_oil_price': max_oil_price,
        'min_gas_price': min_gas_price,
        'mean_gas_price': mean_gas_price,
        'max_gas_price': max_gas_price,
        'min_oil_lifting': min_oil_lifting,
        'mean_oil_lifting': mean_oil_lifting,
        'max_oil_lifting': max_oil_lifting,
        'min_gas_lifting': min_gas_lifting,
        'mean_gas_lifting': mean_gas_lifting,
        'max_gas_lifting': max_gas_lifting,
    }

    # If the element is not being input from argument, fill with original value
    for element in min_max_mean_std.keys():
        if min_max_mean_std[element] is None:
            min_max_mean_std[element] = min_max_mean_original[element]

    # Constructing the dictionary to contain the uncertainty data
    uncertainty_dict = {
        'CAPEX': {
            'min': min_max_mean_std['min_capex'],
            'mean': min_max_mean_std['mean_capex'],
            'max': min_max_mean_std['max_capex'],
            'std_dev': capex_stddev,
            'distribution': capex_distribution,
        },
        'OPEX': {
            'min': min_max_mean_std['min_opex'],
            'mean': min_max_mean_std['mean_opex'],
            'max': min_max_mean_std['max_opex'],
            'std_dev': opex_stddev,
            'distribution': opex_distribution,
        },
        'OILPRICE': {
            'min': min_max_mean_std['min_oil_price'],
            'mean': min_max_mean_std['mean_oil_price'],
            'max': min_max_mean_std['max_oil_price'],
            'std_dev': oil_price_stddev,
            'distribution': oil_price_distribution,
        },
        'GASPRICE': {
            'min': min_max_mean_std['min_gas_price'],
            'mean': min_max_mean_std['mean_gas_price'],
            'max': min_max_mean_std['max_gas_price'],
            'std_dev': gas_price_stddev,
            'distribution': gas_price_distribution,
        },
        'OILLIFTING': {
            'min': min_max_mean_std['min_oil_lifting'],
            'mean': min_max_mean_std['mean_oil_lifting'],
            'max': min_max_mean_std['max_oil_lifting'],
            'std_dev': oil_lifting_stddev,
            'distribution': oil_lifting_distribution,
        },
        'GASLIFTING': {
            'min': min_max_mean_std['min_gas_lifting'],
            'mean': min_max_mean_std['mean_gas_lifting'],
            'max': min_max_mean_std['max_gas_lifting'],
            'std_dev': gas_lifting_stddev,
            'distribution': gas_lifting_distribution,
        },
    }

    # Removing the same value or zeros value of min max and mean element
    uncertainty_dict = {
        key: value for key, value in uncertainty_dict.items() if np.array([value['min'], value['mean'], value['max']]).std() != 0.0
    }

    # Get the multipliers based on the constructed uncertainty dictionary
    for element in uncertainty_dict.keys():
        uncertainty_dict[element]['multipliers'] = get_multipliers_monte(
            run_number=run_number,
            distribution=uncertainty_dict[element]['distribution'],
            min_value=uncertainty_dict[element]['min'],
            mean_value=uncertainty_dict[element]['mean'],
            max_value=uncertainty_dict[element]['max'],
            std_dev=uncertainty_dict[element]['std_dev'],
        )

    # Make a container for the adjusted contract
    adjusted_contract_list = [contract] * run_number

    # Adjust the element of each contract based on each multiplier simultaneously
    for index in range(run_number):
        for element in uncertainty_dict.keys():
            adjusted_contract_list[index] = _adjust_element_single_contract(
                contract=adjusted_contract_list[index],
                contract_arguments=contract_arguments,
                element=element,
                adjustment_value=uncertainty_dict[element]['multipliers'][index],
                run_contract=False,
            )

    ## The alternative code using lambda
    # for index, contract in enumerate(adjusted_contract_list):
    #     adjusted_contract_list[index] = reduce(
    #         lambda psc, element: _adjust_element_single_contract(
    #             contract=psc,
    #             contract_arguments=contract_arguments,
    #             element=element,
    #             adjustment_value=uncertainty_dict[element]['multipliers'][index],
    #             run_contract=False,
    #         ),
    #         uncertainty_dict.keys(),
    #         contract
    #     )

    # Running all contract in the adjusted_contract_list
    for cntrct in adjusted_contract_list:
        cntrct.run(**contract_arguments)

    # Retrieving the summary of each contract in adjusted_contract_list
    summary_list = [get_summary(
                **{**summary_arguments, 'contract': contrct}
            ) for contrct in adjusted_contract_list]

    # Todo: separate the routine of adjust and running the contract to accommodate Multi Processing
    # Constructing a dictionary to contain the sorted result
    target = ["ctr_npv", "ctr_irr", "ctr_pi", "ctr_pot", "gov_take", "ctr_net_share"]
    filtered_result = {
        key :np.array([
            summary[key]
            for summary in summary_list
        ])
        for key in target
    }

    # Sorting the result
    for indicator in filtered_result.keys():
        filtered_result[indicator] = np.take_along_axis(
            arr=filtered_result[indicator],
            indices=np.argsort(filtered_result[indicator], axis=0),
            axis=0,
        )

    percentiles_dict = {
        key: np.percentile(
            a=filtered_result[key],
            q=[10, 50, 90],
            method="lower",
            axis=0,
        )
        for key in filtered_result.keys()
    }

    # Defining the array of probability
    filtered_result['prob'] = (
            np.arange(1, run_number + 1, dtype=np.float_)
            / run_number
    )

    if dataframe_output is True:
        return (pd.DataFrame(filtered_result).set_index('prob'),
                pd.DataFrame(percentiles_dict).set_index(pd.Index(['P10', 'P50', 'P90'])))

    else:
        return (filtered_result,
                percentiles_dict)



from datetime import date

import numpy as np

import pyscnomics.econ.indicator as indicator

cashflow = np.array(
    [
        -100,
        50,
        50,
        50,
        50,
        50,
        50,
        50,
    ]
)
start_date = date(2023, 7, 15)
end_date = date(2030, 7, 14)


def test_pot():
    result = 2.5
    cash = np.array([-100, 50, 40, 20])
    calc = indicator.pot(cashflow=cash)
    np.testing.assert_allclose(calc, result)


def test_npv():
    result = 143.420940884647
    calc = indicator.npv(cashflow=cashflow)
    np.testing.assert_allclose(calc, result)


def test_xnpv():
    result = 154.734614499426  # excel xnpv result
    calc = indicator.xnpv(
        cashflow=cashflow,
        start_date=start_date,
        end_date=end_date,
    )
    np.testing.assert_allclose(result, calc)


def test_irr():
    result = 0.46557123187676  # excel irr result
    calc = indicator.irr(cashflow=cashflow)
    np.testing.assert_allclose(calc, result)


def test_xirr():
    result = 0.621820921  # excel xirr result
    calc = indicator.xirr(
        cashflow=cashflow,
        start_date=start_date,
        end_date=end_date,
    )
    np.testing.assert_allclose(calc, result)



"""
Python Script as the entry point of Excel Workbook
"""

import click
import numpy as np
import pandas as pd
import xlwings as xw
import uvicorn

from pyscnomics.io.parse import InitiateContract
from pyscnomics.io.aggregator import Aggregate
from pyscnomics.io.plot_generator import get_uncertainty_plot

from pyscnomics.tools.summary import get_summary
from pyscnomics.io.write_excel import write_cashflow, write_summary, write_opt, write_table
from pyscnomics.optimize.optimization import optimize_psc

from pyscnomics.contracts.project import BaseProject
from pyscnomics.contracts.costrecovery import CostRecovery
from pyscnomics.contracts.grossplit import GrossSplit
from pyscnomics.contracts.transition import Transition

from pyscnomics.optimize.sensitivity import (
    get_multipliers_sensitivity,
    execute_sensitivity,
)

from pyscnomics.optimize.uncertainty import (
    get_montecarlo_data,
    get_multipliers_montecarlo,
    execute_montecarlo,
)


# Click command for generating CLI command
def main(workbook_path, mode):
    """
    The entry point of the engine if the execution is using Excel workbook.

    Parameters
    ----------
    workbook_path: str
        The path of the workbook
    mode: str
        The mode of the simulation.
        The available mode are: 'Standard', 'Optimization', 'Sensitivity', 'Uncertainty'
    """
    # Defining the workbook object
    workbook_object = xw.Book(workbook_path)

    # Run standard economic analysis
    if mode == "Standard":
        (
            psc,
            psc_arguments,
            summary_arguments,
            data
        ) = InitiateContract(workbook_path=workbook_path).activate()

        run_standard(
            contract=psc,
            contract_arguments=psc_arguments,
            workbook_object=workbook_object,
            summary_argument=summary_arguments,
        )

    # Run optimization study
    elif mode == "Optimization":
        (
            psc,
            psc_arguments,
            summary_arguments,
            data
        ) = InitiateContract(workbook_path=workbook_path).activate()

        run_optimization(
            contract=psc,
            data=data,
            contract_arguments=psc_arguments,
            workbook_object=workbook_object,
            summary_arguments=summary_arguments,
        )

    # Run sensitivity study
    elif mode == "Sensitivity":
        # Prepare the loaded data
        data = Aggregate(workbook_to_read=workbook_path)
        data.fit()

        # Specify the multipliers
        multipliers, total_run = get_multipliers_sensitivity(
            min_deviation=data.sensitivity_data.percentage_min,
            max_deviation=data.sensitivity_data.percentage_max,
            number_of_params=len(data.sensitivity_data.parameter),
        )

        # Run sensitivity study
        target = ["npv", "irr", "pi", "pot", "gov_take", "ctr_net_share"]
        results = execute_sensitivity(
            data=data,
            target=target,
            multipliers=multipliers,
            workbook_path=workbook_path,
        )

        # Arrange the results into the desired output structure
        results_arranged = {
            key: (
                np.zeros(
                    [multipliers.shape[1], len(data.sensitivity_data.parameter)],
                    dtype=np.float_
                )
            )
            for key in target
        }

        for i, econ in enumerate(target):
            for j, param in enumerate(data.sensitivity_data.parameter):
                results_arranged[econ][:, j] = results[param][:, i]

        # Grouping the sensitivity result
        list_df = [pd.DataFrame] * len(target)
        for index, key in enumerate(target):
            df = pd.DataFrame()
            df['oil_price'] = results_arranged[key][:, 0]
            df['gas_price'] = results_arranged[key][:, 1]
            df['opex'] = results_arranged[key][:, 2]
            df['capex'] = results_arranged[key][:, 3]
            df['prod'] = results_arranged[key][:, 4]
            list_df[index] = df

        # Writing the sensitivity result into the workbook
        list_cell_sensitivity = ['M4', 'M29', 'M54', 'M79', 'M104', 'M129']
        for index, cell in enumerate(list_cell_sensitivity):
            write_table(workbook_object=workbook_object,
                        sheet_name='Sensitivity',
                        starting_cell=cell,
                        table=list_df[index], )

    # Run montecarlo simulation
    elif mode == "Uncertainty":
        # Prepare the loaded data
        data = Aggregate(workbook_to_read=workbook_path)
        data.fit()

        # Prepare MonteCarlo data
        uncertainty_data, mean_values_not_zero = get_montecarlo_data(data=data)

        # Get multipliers
        multipliers = np.ones(
            [uncertainty_data["run_number"], len(uncertainty_data["parameter"])],
            dtype=np.float_
        )

        for i in mean_values_not_zero:
            multipliers[:, i] = get_multipliers_montecarlo(
                run_number=uncertainty_data["run_number"],
                distribution=uncertainty_data["distribution"][i],
                min_value=uncertainty_data["min_values"][i],
                mean_value=uncertainty_data["mean_values"][i],
                max_value=uncertainty_data["max_values"][i],
                std_dev=uncertainty_data["std_dev"][i],
            )

        # Run MonteCarlo simulation
        target = ["npv", "irr", "pi", "pot", "gov_take", "ctr_net_share"]
        results = execute_montecarlo(
            data=data,
            workbook_path=workbook_path,
            uncertainty_data=uncertainty_data,
            target=target,
            mult=multipliers,
        )

        # Sorted the results
        results_sorted = np.take_along_axis(
            arr=results,
            indices=np.argsort(results, axis=0),
            axis=0,
        )

        # Specify probability
        prob = (
                np.arange(1, uncertainty_data["run_number"] + 1, dtype=np.float_)
                / uncertainty_data["run_number"]
        )

        # Arrange the results
        results_arranged = np.concatenate((prob[:, np.newaxis], results_sorted), axis=1)

        # Calculate P10, P50, P90
        percentiles = np.percentile(
            a=results_arranged,
            q=[10, 50, 90],
            method="lower",
            axis=0,
        )

        # Determine indices of data to be captured to Excel
        indices = (
            np.linspace(0, uncertainty_data["run_number"], 101)[0:-1].astype(int)
        )

        # Final outcomes to be captured to Excel
        outcomes = {
            "results": results_arranged[indices, :],
            "P10": percentiles[0, :],
            "P50": percentiles[1, :],
            "P90": percentiles[2, :],
        }

        # Making the dataframe to contain the result of percentile
        df_uncertainty_percentile = pd.DataFrame()
        df_uncertainty_percentile.index = ['Result', 'NPV', 'IRR', 'PI', 'POT', 'Gov_Take', 'CTR_Net_Share']
        df_uncertainty_percentile['P10'] = outcomes['P10']
        df_uncertainty_percentile['P50'] = outcomes['P50']
        df_uncertainty_percentile['P90'] = outcomes['P90']
        df_uncertainty_transposed = df_uncertainty_percentile.transpose()
        df_uncertainty_transposed.drop(['Result'], axis=1, inplace=True)

        # Grouping the data into each category in a dataframe
        df_uncertainty_result = pd.DataFrame()
        df_uncertainty_result['frequency'] = outcomes['results'][:, 0]
        df_uncertainty_result['npv'] = outcomes['results'][:, 1]
        df_uncertainty_result['irr'] = outcomes['results'][:, 2]
        df_uncertainty_result['pi'] = outcomes['results'][:, 3]
        df_uncertainty_result['pot'] = outcomes['results'][:, 4]
        df_uncertainty_result['gov_take'] = outcomes['results'][:, 5]
        df_uncertainty_result['ctr_net_share'] = outcomes['results'][:, 6]

        # Writing the percentile table into Excel workbook
        write_table(
            workbook_object=workbook_object,
            sheet_name='Uncertainty',
            starting_cell='L6',
            table=df_uncertainty_transposed,
        )

        # Writing the result table into Excel workbook
        write_table(
            workbook_object=workbook_object,
            sheet_name='Uncertainty',
            starting_cell='K49',
            table=df_uncertainty_result,
        )

        # Generating the uncertainty plot
        get_uncertainty_plot(uncertainty_outcomes=outcomes, plot_type='Stairway')

    # Giving the workbook execution status to show that execution is success
    xw.Book(workbook_path).sheets("References").range("N17").value = "Success"


@click.command()
@click.option(
    '-p',
    '--path',
    help='The path of the Microsoft Excel Workbook with PySCnomics template'
)
@click.option(
    '-m',
    '--mode',
    default='Standard',
    help='The mode of the simulation. They are: "Standard", "Sensitivity", "Optimization", "Uncertainty"'
)
@click.option(
    '-api',
    '--api',
    default=1,
    help='The command for running the API backend. '
         '0 for not activating the API backend. 1 for activating the API backend'
)
@click.option(
    '-port',
    '--port',
    default=8000,
    help='The port number for running the API backend. The default port is 8000'
)
def entry_point(**kwargs):
    """ Manages CLI """
    if kwargs['api'] == 1:
        body = """
                We welcome you to our library, PySCnomics. This package contains tailored functionalities for 
                assessing economic feasibility of oil and gas projects following the state-of-the-art Production 
                Sharing Contract (PSC) schemes in Indonesia.
                PySCnomics is the product of join research between Indonesia's Special Task Force for Upstream Oil 
                and Gas Business Activities (SKK Migas) and the Department of Petroleum Engineering, 
                Institut Teknologi Bandung (ITB)
                """
        print(body)
        port_number = kwargs['port']
        uvicorn.run("pyscnomics.api.main:app", host="0.0.0.0", port=int(port_number), reload=False)

    if kwargs['path'] is not None:
        # Defining the cli command for path
        if 'p' in kwargs:
            file_path = kwargs['p']
        elif 'path' in kwargs:
            file_path = kwargs['path']
        else:
            file_path = None

        # Defining the cli command for mode
        if 'm' in kwargs:
            mode = kwargs['m']
        elif 'mode' in kwargs:
            mode = kwargs['mode']
        else:
            mode = None

        # Running the code based on the given CLI input
        main(workbook_path=file_path, mode=mode)


def run_standard(
        contract: CostRecovery | GrossSplit | Transition,
        contract_arguments: dict,
        workbook_object: xw.Book,
        summary_argument: dict
):
    """
    The function to run simulation in Standard mode.

    Parameters
    ----------
    contract: CostRecovery | GrossSplit | Transition
        The contract object that will be run
    contract_arguments: dict
        The contract arguments
    workbook_object: xw.Book
        The workbook object used in simulation
    summary_argument: dict
        The dictionary of the summary arguments
    """
    # Running standard contract
    contract.run(**contract_arguments)

    # Define the summary object
    contract_summary = get_summary(**summary_argument)

    sheet_name = None
    if isinstance(contract, CostRecovery):
        sheet_name = 'Result Table CR'

    elif isinstance(contract, GrossSplit):
        sheet_name = 'Result Table GS'

    elif isinstance(contract, BaseProject):
        sheet_name = 'Result Table Base Project'

    elif isinstance(contract, Transition):
        sheet_name = 'Transition'

    # Writing the result of the contract
    write_cashflow(
        workbook_object=workbook_object,
        sheet_name=sheet_name,
        contract=contract,
    )

    # Writing the summary of the contract
    write_summary(
        summary_dict=contract_summary,
        workbook_object=workbook_object,
        sheet_name='Executive Summary',
        range_cell='E5',
    )


def run_optimization(
        contract: CostRecovery | GrossSplit | Transition,
        data: Aggregate,
        contract_arguments: dict,
        summary_arguments: dict,
        workbook_object: xw.Book,
):
    """
    The function to run simulation in Optimization mode.

    Parameters
    ----------
    """
    # Running standard contract
    contract.run(**contract_arguments)

    # Condition to adjust the dict_optimization based on the contract type
    if isinstance(contract, CostRecovery):
        dict_optimization = data.optimization_data.data_cr
        range_list_params = "N5"
        range_list_value = "P5"

    elif isinstance(contract, GrossSplit):
        dict_optimization = data.optimization_data.data_gs
        range_list_params = "N17"
        range_list_value = "P17"

    else:
        if isinstance(contract.contract2, CostRecovery):
            dict_optimization = data.optimization_data.data_cr
            range_list_params = "N5"
            range_list_value = "P5"
        else:
            dict_optimization = data.optimization_data.data_gs
            range_list_params = "N17"
            range_list_value = "P17"

    # Defining the target_optimization_value and target_parameter
    target_optimization_value = data.optimization_data.target["parameter"]
    target_parameter = data.optimization_data.target["value"]

    # Running optimization
    optim_result = optimize_psc(
        dict_optimization=dict_optimization,
        contract=contract,
        contract_arguments=contract_arguments,
        target_optimization_value=target_parameter,
        summary_argument=summary_arguments,
        target_parameter=target_optimization_value,
    )

    list_str = optim_result[0]
    list_params_value = optim_result[1]
    result_optimization = optim_result[2]
    optimized_contract = optim_result[3][-1]

    # Writing optimization result
    write_opt(
        list_str=list_str,
        list_params_value=list_params_value,
        result_optimization=result_optimization,
        workbook_object=workbook_object,
        sheet_name='Optimization',
        range_opt_result='P2',
        range_list_params=range_list_params,
        range_list_value=range_list_value,
    )

    # Generating the summary of the optimized contract
    summary_arguments['contract'] = optimized_contract
    optimized_contract_summary = get_summary(**summary_arguments)

    # Writing optimized contract summary
    write_summary(
        summary_dict=optimized_contract_summary,
        workbook_object=workbook_object,
        sheet_name='Optimization Comparison',
        range_cell='J6',
    )


if __name__ == "__main__":
    entry_point()



"""
A collection of unit testing for CostOfSales class.
"""

import pytest
import numpy as np
from pyscnomics.econ.selection import FluidType
from pyscnomics.econ.costs import CostOfSales, CostOfSalesException


def test_cos_incorrect_start_year_end_year():
    """ A unit testing for incorrect data input: start year is after end year """

    with pytest.raises(CostOfSalesException):
        CostOfSales(
            start_year=2031,
            end_year=2030,
        )


def test_cos_incorrect_expense_year():
    """
    A unit testing for incorrect expense year data input:
    expense_year data is not given as a numpy.ndarray datatype.
    """
    with pytest.raises(CostOfSalesException):
        CostOfSales(
            start_year=2023,
            end_year=2030,
            expense_year=[2023, 2024, 2029],
        )


def test_cos_incorrect_cost():
    """
    A unit testing for incorrect cost data input:
    cost data is not given as a numpy.ndarray datatype.
    """
    with pytest.raises(CostOfSalesException):
        CostOfSales(
            start_year=2023,
            end_year=2030,
            expense_year=np.array([2023, 2024, 2025, 2026]),
            cost=[50, 50, 50, 50],
        )


def test_cos_incorrect_cost_allocation():
    """
    A unit testing for incorrect cost_allocation data input:
    cost_allocation is not given as a list datatype.
    """
    with pytest.raises(CostOfSalesException):
        CostOfSales(
            start_year=2023,
            end_year=2030,
            expense_year=np.array([2023, 2024, 2029]),
            cost=np.array([150, 150, 100]),
            cost_allocation=FluidType.GAS,
        )


def test_cos_incorrect_unequal_length_array():
    """ A unit testing for unequal length of arrays """

    with pytest.raises(CostOfSalesException):
        CostOfSales(
            start_year=2023,
            end_year=2030,
            expense_year=np.array([2023, 2024, 2029]),
            cost=np.array([50, 50, 50, 50]),
            cost_allocation=[FluidType.GAS, FluidType.GAS, FluidType.GAS],
        )


def test_cos_incorrect_year_input():
    """
    A unit testing for incorrect expense year data input:
    (1) Expense year is before the start year of the project,
    (2) Expense year is after the end year of the project
    """
    with pytest.raises(CostOfSalesException):
        CostOfSales(
            start_year=2023,
            end_year=2030,
            expense_year=np.array([2019, 2024, 2025]),
        )

    with pytest.raises(CostOfSalesException):
        CostOfSales(
            start_year=2023,
            end_year=2030,
            expense_year=np.array([2024, 2025, 2032]),
        )


def test_cos_prepare_expense_year():
    """
    Test the preparation of `expense_year` attribute in the `CostOfSales` class.

    This function tests whether the `expense_year` attribute is correctly initialized
    in the `CostOfSales` class. Two test cases are checked:

    1. When `expense_year` is not provided, it should be initialized to the range
       from `start_year` to `end_year`.
    2. When `expense_year` is provided, it should be set correctly in the instance.

    Expected Results
    ----------------
    expense_year_exp1 : np.ndarray
        The expected `expense_year` array when it is not provided.
    expense_year_exp2 : np.ndarray
        The expected `expense_year` array when it is provided.

    Calculated Results
    ------------------
    cos1 : CostOfSales
        An instance of `CostOfSales` with default `expense_year`.
    cos2 : CostOfSales
        An instance of `CostOfSales` with a specified `expense_year`.
    expense_year_calc1 : np.ndarray
        The calculated `expense_year` array for `cos1`.
    expense_year_calc2 : np.ndarray
        The calculated `expense_year` array for `cos2`.

    Assertions
    ----------
    np.testing.assert_allclose(expense_year_exp1, expense_year_calc1)
        Check if the expected `expense_year` matches the calculated value for `cos1`.
    np.testing.assert_allclose(expense_year_exp2, expense_year_calc2)
        Check if the expected `expense_year` matches the calculated value for `cos2`.
    """

    # Expected results
    expense_year_exp1 = np.array([2023, 2024, 2025, 2026, 2027, 2028, 2029, 2030])
    expense_year_exp2 = np.array([2023, 2024, 2029])

    # Calculated results
    cos1 = CostOfSales(
        start_year=2023,
        end_year=2030,
    )

    cos2 = CostOfSales(
        start_year=2023,
        end_year=2030,
        expense_year=np.array([2023, 2024, 2029]),
        cost=np.array([150, 150, 100]),
        cost_allocation=[FluidType.OIL, FluidType.OIL, FluidType.OIL],
    )

    expense_year_calc1 = cos1.expense_year
    expense_year_calc2 = cos2.expense_year

    # Check whether expected == calculated
    np.testing.assert_allclose(expense_year_exp1, expense_year_calc1)
    np.testing.assert_allclose(expense_year_exp2, expense_year_calc2)


def test_cos_prepare_cost():
    """
    Test the preparation of the `cost` attribute in the `CostOfSales` class.

    This function tests whether the `cost` attribute is correctly initialized in the
    `CostOfSales` class. Two test cases are checked:

    1. When `cost` is not provided, it should be initialized to an array of zeros
       with the same length as the project duration.
    2. When `cost` is provided, it should be set correctly in the instance.

    Expected Results
    ----------------
    cost_exp1 : np.ndarray
        The expected `cost` array when it is not provided.
    cost_exp2 : np.ndarray
        The expected `cost` array when it is provided.

    Calculated Results
    ------------------
    cos1 : CostOfSales
        An instance of `CostOfSales` with default `cost`.
    cos2 : CostOfSales
        An instance of `CostOfSales` with a specified `cost`.
    cost_calc1 : np.ndarray
        The calculated `cost` array for `cos1`.
    cost_calc2 : np.ndarray
        The calculated `cost` array for `cos2`.

    Assertions
    ----------
    np.testing.assert_allclose(cost_exp1, cost_calc1)
        Check if the expected `cost` matches the calculated value for `cos1`.
    np.testing.assert_allclose(cost_exp2, cost_calc2)
        Check if the expected `cost` matches the calculated value for `cos2`.
    """

    # Expected results
    cost_exp1 = np.array([0., 0., 0., 0., 0., 0., 0., 0.])
    cost_exp2 = np.array([150, 150, 100])

    # Calculated results
    cos1 = CostOfSales(
        start_year=2023,
        end_year=2030,
    )

    cos2 = CostOfSales(
        start_year=2023,
        end_year=2030,
        expense_year=np.array([2024, 2025, 2029]),
        cost=np.array([150, 150, 100]),
        cost_allocation=[FluidType.OIL, FluidType.OIL, FluidType.OIL],
    )

    cost_calc1 = cos1.cost
    cost_calc2 = cos2.cost

    # Check whether expected == calculated
    np.testing.assert_allclose(cost_exp1, cost_calc1)
    np.testing.assert_allclose(cost_exp2, cost_calc2)


def test_cos_prepare_cost_allocation():
    """
    Test the preparation of the `cost_allocation` attribute in the `CostOfSales` class.

    This function tests whether the `cost_allocation` attribute is correctly initialized
    in the `CostOfSales` class. Two test cases are checked:

    1. When `cost_allocation` is not provided, it should be initialized to a list of
       `FluidType.OIL` with the same length as the project duration.
    2. When `cost_allocation` is provided, it should be set correctly in the instance.

    Expected Results
    ----------------
    cost_allocation_exp1 : list of FluidType
        The expected `cost_allocation` list when it is not provided.
    cost_allocation_exp2 : list of FluidType
        The expected `cost_allocation` list when it is provided.

    Calculated Results
    ------------------
    cos1 : CostOfSales
        An instance of `CostOfSales` with default `cost_allocation`.
    cos2 : CostOfSales
        An instance of `CostOfSales` with a specified `cost_allocation`.
    cost_allocation_calc1 : list of FluidType
        The calculated `cost_allocation` list for `cos1`.
    cost_allocation_calc2 : list of FluidType
        The calculated `cost_allocation` list for `cos2`.

    Assertions
    ----------
    assert cost_allocation_exp1 == cost_allocation_calc1
        Check if the expected `cost_allocation` matches the calculated value for `cos1`.
    assert cost_allocation_exp2 == cost_allocation_calc2
        Check if the expected `cost_allocation` matches the calculated value for `cos2`.
    """
    # Expected results
    cost_allocation_exp1 = [
        FluidType.OIL, FluidType.OIL, FluidType.OIL, FluidType.OIL,
        FluidType.OIL, FluidType.OIL, FluidType.OIL, FluidType.OIL,
    ]

    cost_allocation_exp2 = [
        FluidType.OIL, FluidType.OIL, FluidType.GAS, FluidType.GAS,
    ]

    # Calculated results
    cos1 = CostOfSales(
        start_year=2023,
        end_year=2030,
    )

    cos2 = CostOfSales(
        start_year=2023,
        end_year=2030,
        expense_year=np.array([2023, 2024, 2025, 2026]),
        cost=np.array([50, 50, 50, 50]),
        cost_allocation=[FluidType.OIL, FluidType.OIL, FluidType.GAS, FluidType.GAS]
    )

    cost_allocation_calc1 = cos1.cost_allocation
    cost_allocation_calc2 = cos2.cost_allocation

    # Check whether expected == calculated
    assert cost_allocation_exp1 == cost_allocation_calc1
    assert cost_allocation_exp2 == cost_allocation_calc2



"""
A collection of unit testing for Tangible class
"""

import pytest
import numpy as np
from pyscnomics.econ.selection import DeprMethod, FluidType, TaxType
from pyscnomics.econ.costs import CapitalException, CapitalCost


def test_capital_incorrect_year_input():
    """A unit testing for incorrect data input: start year is after end year"""

    with pytest.raises(CapitalException):
        CapitalCost(
            start_year=2030,
            end_year=2023,
            cost=np.array([200, 200]),
            expense_year=np.array([2023, 2024]),
            salvage_value=np.array([25, 25]),
            useful_life=np.array([5, 5]),
            cost_allocation=[FluidType.OIL, FluidType.OIL],
        )


def test_capital_unequal_length_of_data_input():
    """A unit testing for incorrect data input: unequal length of data input"""

    with pytest.raises(CapitalException):
        CapitalCost(
            start_year=2023,
            end_year=2032,
            cost=np.array([200, 200]),
            expense_year=np.array([2023, 2024, 2025]),
            salvage_value=np.array([25, 25]),
            useful_life=np.array([5, 5]),
            cost_allocation=[FluidType.OIL, FluidType.OIL],
        )


def test_capital_incorrect_expense_year_input():
    """A unit testing for incorrect data input: expense year is after end year of the project"""

    with pytest.raises(CapitalException):
        CapitalCost(
            start_year=2023,
            end_year=2032,
            cost=np.array([200, 200]),
            expense_year=np.array([2023, 2040]),
            salvage_value=np.array([25, 25]),
            useful_life=np.array([5, 5]),
            cost_allocation=[FluidType.OIL, FluidType.OIL],
        )

    with pytest.raises(CapitalException):
        CapitalCost(
            start_year=2023,
            end_year=2032,
            cost=np.array([200, 200]),
            expense_year=np.array([2020, 2023]),
            cost_allocation=[FluidType.OIL, FluidType.OIL],
        )


def test_capital_comparison():
    """Units testing for comparison operation"""

    mangga_capital = CapitalCost(
        start_year=2023,
        end_year=2030,
        cost=np.array([200, 200]),
        expense_year=np.array([2023, 2024]),
        salvage_value=np.array([25, 25]),
        useful_life=np.array([5, 5]),
        cost_allocation=[FluidType.OIL, FluidType.OIL],
    )

    jeruk_capital = CapitalCost(
        start_year=2023,
        end_year=2030,
        cost=np.array([200, 200]),
        expense_year=np.array([2023, 2024]),
        salvage_value=np.array([25, 25]),
        useful_life=np.array([5, 5]),
        cost_allocation=[FluidType.OIL, FluidType.OIL],
    )

    hiu_capital = CapitalCost(
        start_year=2025,
        end_year=2030,
        cost=np.array([100, 100]),
        expense_year=np.array([2025, 2026]),
        cost_allocation=[FluidType.GAS, FluidType.GAS],
    )

    # Execute comparison operations
    assert mangga_capital == jeruk_capital
    assert mangga_capital != hiu_capital
    assert mangga_capital == 400
    assert mangga_capital >= hiu_capital
    assert hiu_capital <= jeruk_capital


def test_capital_arithmetics_misuse():
    """A unit testing for incorrect arithmetic operations"""

    mangga_capital = CapitalCost(
        start_year=2023,
        end_year=2030,
        cost=np.array([100, 100]),
        expense_year=np.array([2023, 2024]),
        cost_allocation=[FluidType.OIL, FluidType.OIL],
    )

    with pytest.raises(CapitalException):
        assert mangga_capital + 500

    with pytest.raises(CapitalException):
        assert mangga_capital - 100

    with pytest.raises(CapitalException):
        assert mangga_capital / 0


def test_capital_arithmetics():
    """A unit testing for mathematical operations (associated with an instance of Capital class)"""

    mangga_capital = CapitalCost(
        start_year=2023,
        end_year=2030,
        cost=np.array([100, 100]),
        expense_year=np.array([2024, 2023]),
        cost_allocation=[FluidType.OIL, FluidType.OIL],
    )

    apel_capital = CapitalCost(
        start_year=2023,
        end_year=2030,
        cost=np.array([50, 50]),
        expense_year=np.array([2026, 2027]),
        cost_allocation=[FluidType.OIL, FluidType.OIL],
    )

    jeruk_capital = CapitalCost(
        start_year=2023,
        end_year=2032,
        cost=np.array([200]),
        expense_year=np.array([2023]),
        cost_allocation=[FluidType.OIL],
    )

    # Expected results
    add1 = np.array([100, 100, 50, 50])
    add2 = np.array([100, 100, 50, 50, 200])
    mul1 = np.array([200, 200])
    mul2 = np.array([-25, -25])
    div1 = np.array([10])
    div2 = 0.5

    # Calculated results
    calc_add1 = mangga_capital + apel_capital
    calc_add2 = mangga_capital + apel_capital + jeruk_capital
    calc_mul1 = mangga_capital * 2
    calc_mul2 = -0.5 * apel_capital
    calc_div1 = jeruk_capital / 20
    calc_div2 = apel_capital / mangga_capital

    # Execute tests
    np.testing.assert_allclose(add1, calc_add1.cost)
    np.testing.assert_allclose(add2, calc_add2.cost)
    np.testing.assert_allclose(mul1, calc_mul1.cost)
    np.testing.assert_allclose(mul2, calc_mul2.cost)
    np.testing.assert_allclose(div1, calc_div1.cost)
    assert div2 == calc_div2


def test_capital_expenditures():
    """A unit testing for expenditures method in Capital class"""

    # Expected result
    expenses = [100, 100, 150, 0, 0, 0, 0, 0, 0, 0]

    # Calculated result
    mangga_capital = CapitalCost(
        start_year=2023,
        end_year=2030,
        cost=np.array([100, 100, 100]),
        expense_year=np.array([2023, 2024, 2025]),
        useful_life=np.array([5, 5, 5]),
        cost_allocation=[FluidType.OIL, FluidType.OIL, FluidType.OIL],
    )

    jeruk_capital = CapitalCost(
        start_year=2025,
        end_year=2032,
        cost=np.array([50]),
        expense_year=np.array([2025]),
        useful_life=np.array([5]),
        cost_allocation=[FluidType.OIL],
    )

    total_capital = mangga_capital + jeruk_capital
    calc_expenses = total_capital.expenditures_post_tax()

    # Execute testing (expected == calculated)
    np.testing.assert_allclose(expenses, calc_expenses)


def test_capital_expenditures_with_tax_and_inflation():
    """
    A unit testing for expenditures method in Capital class,
    taking into account tax and inflation schemes.
    """

    # Expected results
    case1 = np.array([0, 102, 104.04, 106.1208, 108.243216, 0, 0, 0])
    case2 = np.array([0, 102, 105.06, 109.2624, 114.72552, 0, 0, 0])
    case3 = np.array([0, 112, 112, 112, 112, 0, 0, 0])

    # Calculated results
    mangga_capital = CapitalCost(
        start_year=2023,
        end_year=2030,
        cost=np.array([100, 100, 100, 50]),
        expense_year=np.array([2024, 2025, 2026, 2027]),
        cost_allocation=[FluidType.OIL, FluidType.OIL, FluidType.OIL, FluidType.OIL],
        tax_portion=np.array([1, 1, 1, 1]),
    )

    jeruk_capital = CapitalCost(
        start_year=2023,
        end_year=2030,
        cost=np.array([50]),
        expense_year=np.array([2027]),
        cost_allocation=[FluidType.OIL],
        tax_portion=np.array([1,]),
    )

    capital_add = mangga_capital + jeruk_capital

    case1_calc = capital_add.expenditures_post_tax(
        inflation_rate=0.02
    )
    case2_calc = capital_add.expenditures_post_tax(
        inflation_rate=np.array([0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.07, 0.08]),
    )
    case3_calc = capital_add.expenditures_post_tax(
        tax_rate=0.12,
    )

    # Execute testing (expected == calculated)
    np.testing.assert_allclose(case1, case1_calc)
    np.testing.assert_allclose(case2, case2_calc)
    np.testing.assert_allclose(case3, case3_calc)


def test_capital_depreciation():
    """A unit testing for total depreciation rate of Capital object"""

    # Expected results: depreciation_charge, undepreciated asset
    depreSL = [0, 0, 20, 30, 30, 30, 30, 10]
    depreDB = [0, 0, 20, 26, 20.8, 16.64, 46.08, 20.48]
    depreDDB = [0, 0, 40, 24, 14.4, 28.64, 24.96, 7.2]
    deprePSC = [0, 0, 50, 25, 12.5, 6.25, 31.25, 12.5]

    undepreSL = 0
    undepreDB = 0
    undepreDDB = 10.8
    undeprePSC = 12.5

    # Calculated results: depreciation_charge, undepreciated asset
    mangga_capital = CapitalCost(
        start_year=2023,
        end_year=2030,
        cost=np.array([100, 50]),
        expense_year=np.array([2025, 2026]),
        cost_allocation=[FluidType.OIL, FluidType.OIL],
        useful_life=np.array([5, 5]),
    )

    apel_capital = CapitalCost(
        start_year=2023,
        end_year=2030,
        cost=np.array([100, 50]),
        expense_year=np.array([2025, 2028]),
        cost_allocation=[FluidType.OIL, FluidType.OIL],
        useful_life=np.array([5, 5]),
    )

    nanas_capital = CapitalCost(
        start_year=2023,
        end_year=2030,
        cost=np.array([100, 50]),
        expense_year=np.array([2025, 2029]),
        cost_allocation=[FluidType.OIL, FluidType.OIL],
        useful_life=np.array([5, 5]),
    )

    calc_depreSL, calc_undepreSL = mangga_capital.total_depreciation_rate(
        depr_method=DeprMethod.SL
    )

    calc_depreDB, calc_undepreDB = mangga_capital.total_depreciation_rate(
        depr_method=DeprMethod.DB, decline_factor=1
    )

    calc_depreDDB, calc_undepreDDB = apel_capital.total_depreciation_rate(
        depr_method=DeprMethod.DB, decline_factor=2
    )

    calc_deprePSC, calc_undeprePSC = nanas_capital.total_depreciation_rate(
        depr_method=DeprMethod.PSC_DB
    )

    # Execute testing (expected == calculated)
    np.testing.assert_allclose(depreSL, calc_depreSL)
    np.testing.assert_allclose(depreDB, calc_depreDB)
    np.testing.assert_allclose(depreDDB, calc_depreDDB)
    np.testing.assert_allclose(deprePSC, calc_deprePSC)

    np.testing.assert_allclose(np.array([undepreSL]), np.array([calc_undepreSL]))
    np.testing.assert_allclose(np.array([undepreDB]), np.array([calc_undepreDB]))
    np.testing.assert_allclose(np.array([undepreDDB]), np.array([calc_undepreDDB]))
    np.testing.assert_allclose(np.array([undeprePSC]), np.array([calc_undeprePSC]))


def test_capital_depreciation_with_tax_and_inflation():
    """A unit testing for total depreciation rate of Capital object"""

    # Expected results: depreciation_charge, undepreciated asset
    depreCase1 = [
        50,
        77.5,
        93.875,
        75.878125,
        71.45171875,
        35.88210938,
        18.10511719,
        7.416035156,
    ]

    depreCase2 = [
        50,
        76,
        90.53,
        72.5806,
        68.09668,
        34.11084,
        17.151045,
        6.9996225,
    ]

    undepreCase1 = 3.798457031
    undepreCase2 = 3.58517249999

    # Calculated results
    capital_mangga = CapitalCost(
        start_year=2023,
        end_year=2030,
        cost=np.array([100, 100, 100, 50, 50, 0, 0, 0]),
        expense_year=np.array([2023, 2024, 2025, 2026, 2027, 2028, 2029, 2030]),
        cost_allocation=[
            FluidType.OIL,
            FluidType.OIL,
            FluidType.OIL,
            FluidType.OIL,
            FluidType.OIL,
            FluidType.OIL,
            FluidType.OIL,
            FluidType.OIL,
        ],
    )

    depreCase1_calc, undepreCase1_calc = capital_mangga.total_depreciation_rate(
        inflation_rate=0.05
    )

    depreCase2_calc, undepreCase2_calc = capital_mangga.total_depreciation_rate(
        tax_rate=np.array([0.08, 0.07, 0.06, 0.05, 0.04, 0.03, 0.02, 0.01]),
        inflation_rate=np.array([0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.07, 0.08]),
    )

    # Execute testing
    np.testing.assert_allclose(depreCase1, depreCase1_calc)
    np.testing.assert_allclose(undepreCase1, undepreCase1_calc)
    np.testing.assert_allclose(depreCase2, depreCase2_calc)
    np.testing.assert_allclose(undepreCase2, undepreCase2_calc)


def test_capital_book_value():
    """A unit testing for total depreciation book value of Capital object"""

    # Expected results: depreciation book value
    bookSL = [0, 0, 80, 100, 70, 40, 10, 0]
    bookDB = [0, 0, 80, 104, 83.2, 66.56, 20.48, 0]
    bookDDB = [0, 0, 60, 36, 21.6, 42.96, 18, 10.8]
    bookPSC = [0, 0, 50, 25, 12.5, 6.25, 25, 12.5]

    # Calculated results: depreciation book value
    mangga_capital = CapitalCost(
        start_year=2023,
        end_year=2030,
        cost=np.array([100, 50]),
        expense_year=np.array([2025, 2026]),
        useful_life=np.array([5, 5]),
        cost_allocation=[FluidType.OIL, FluidType.OIL],
    )

    apel_capital = CapitalCost(
        start_year=2023,
        end_year=2030,
        cost=np.array([100, 50]),
        expense_year=np.array([2025, 2028]),
        useful_life=np.array([5, 5]),
        cost_allocation=[FluidType.OIL, FluidType.OIL],
    )

    nanas_capital = CapitalCost(
        start_year=2023,
        end_year=2030,
        cost=np.array([100, 50]),
        expense_year=np.array([2025, 2029]),
        useful_life=np.array([5, 5]),
        cost_allocation=[FluidType.OIL, FluidType.OIL],
    )

    calc_bookSL = mangga_capital.total_depreciation_book_value(
        depr_method=DeprMethod.SL
    )

    calc_bookDB = mangga_capital.total_depreciation_book_value(
        depr_method=DeprMethod.DB, decline_factor=1
    )

    calc_bookDDB = apel_capital.total_depreciation_book_value(
        depr_method=DeprMethod.DB, decline_factor=2
    )

    calc_bookPSC = nanas_capital.total_depreciation_book_value(
        depr_method=DeprMethod.PSC_DB
    )

    # Execute testing
    np.testing.assert_allclose(bookSL, calc_bookSL)
    np.testing.assert_allclose(bookDB, calc_bookDB)
    np.testing.assert_allclose(bookDDB, calc_bookDDB)
    np.testing.assert_allclose(bookPSC, calc_bookPSC)



"""
Collection of functions to carry out some operations related to module costs.py
"""

import numpy as np


class TaxInflationException(Exception):
    """ Exception to be raised for an incorrect tax and inflation configurations """

    pass


def check_input(target_func, param: np.ndarray | float | int) -> np.ndarray:
    """
    Check and prepare input parameters for subsequent analysis.

    Parameters
    ----------
    target_func: Function
        A target function.

    param : numpy.ndarray or float or int
        Parameter to be checked and prepared for analysis.

    Returns
    -------
    numpy.ndarray
        A properly formatted parameter array for analysis.

    Notes:
    ------
    This function checks the input parameter 'param' and ensures it is properly
    prepared for the subsequent analysis.
    If 'param' is a float or int, it creates an array of the same length as
    'target_func' with all elements set to 'param'.
    If 'param' is a numpy.ndarray, it checks that its length matches the length
    of 'target_func'. If not, it raises an exception.
    """

    if not isinstance(param, (float, int, np.ndarray)):
        raise TaxInflationException(
            f"Input parameter must be given as a float, int, or numpy.ndarray, "
            f"not as a/an {param.__class__.__qualname__}"
        )

    else:
        if isinstance(param, np.ndarray):
            if len(param) != len(target_func):
                raise TaxInflationException(
                    f"Unequal length of arrays: "
                    f"{param.__class__.__qualname__}({param}): ({len(param)}), "
                    f"{target_func.__class__.__qualname__}({target_func}): ({len(target_func)})."
                )
            param_arr = param

        elif isinstance(param, (float, int)):
            param_arr = np.repeat(param, len(target_func))

    return param_arr


def get_cost_adjustment_by_tax(
    start_year: int,
    cost: np.ndarray,
    expense_year: np.ndarray,
    project_years: np.ndarray,
    tax_portion: np.ndarray,
    tax_rate: np.ndarray | float,
    tax_discount: float,
) -> np.ndarray:
    """
    Adjust cost based on tax scheme over the prescribed time period.

    Parameters
    ----------
    start_year : int
        The start year of the project.
    cost : np.ndarray
        An array of costs to be adjusted.
    expense_year : np.ndarray
        An array specifying the expense years for each cost element.
    project_years : np.ndarray
        An array of project years.
    tax_portion: np.ndarray
        The portion of 'cost' that is subject to tax.
        Must be an array of length equals to the length of 'cost' array.
    tax_rate: np.ndarray | float
        The tax rate to apply. Can be a single value or an array.
    tax_discount: float
        The tax discount to apply.

    Returns
    -------
    cost_adjusted: np.ndarray
        An array of adjusted costs after applying tax scheme.

    Notes
    -----
    The core operations are as follows:
    (1) Create 'tax_rate_arr' by calling 'check_input()' function. If 'tax_rate'
        is given as an array, the length of the array must be consistent with
        the duration of the project. If it is given as a single value, then create
        an array of length equal to the project duration with all elements set equal
        to the single value,
    (2) Identify the index location of 'tax_rate' based on the associated 'expense_year',
    (3) Calculate the multiplier:
        mult = 1.0 + tax_portion * tax_rate_arr[tax_id] * (1.0 - tax_discount),
    (4) Apply cost adjustment = cost * mult.
    """

    tax_rate_arr = check_input(target_func=project_years, param=tax_rate)
    tax_rate_id = (expense_year - start_year).astype("int")

    return cost * (
        1.0 + tax_portion * tax_rate_arr[tax_rate_id] * (1.0 - tax_discount)
    )


def get_cost_adjustment_by_inflation(
    start_year: int,
    end_year: int,
    cost: np.ndarray,
    expense_year: np.ndarray,
    project_years: np.ndarray,
    year_inflation: np.ndarray,
    inflation_rate: np.ndarray | float,
) -> np.ndarray:
    """
    Adjusts costs for inflation over the project years.

    This function applies inflation adjustments to costs based on the specified years
    and inflation rates. It calculates a multiplier for each cost value, adjusting it
    by inflation rates across the period between `start_year` and the `expense_year`.

    Parameters
    ----------
    start_year : int
        The start year of the project.
    end_year: int
        The end year of the project.
    cost : np.ndarray
        The array of costs to be adjusted for inflation.
    expense_year : np.ndarray
        The year(s) when the expenses occur, corresponding to the provided costs.
    project_years : np.ndarray
        An array of years representing the project timeline for inflation adjustments.
    year_inflation : np.ndarray
        An array of reference year for inflation calculation.
    inflation_rate : np.ndarray or float
        The inflation rate(s) for the project period. If a single float is provided,
        it applies uniformly across all years; otherwise, an array of rates for
        each year should be given.

    Returns
    -------
    np.ndarray
        An array of costs adjusted for inflation over the specified project years.

    Notes
    -----
    The core operations are as follows:
    (1) Check 'inflation_rate' whether it is provided as an array or as a single value.
        If it is given as an array, the length of the array must be consistent with the
        duration of the project. If it is given as a single value, then create an array
        of length equal to project duration with all elements set equal to the single
        value.
    (2) Parameter 'id_start' identify the index location of 'year_inflation' in array
        'project_years'. The result is then added by unity. This parameter sets up
        the first index to slice the 'inflation_rate_arr',
    (3) Parameter 'id_end' configure the index location of 'expense_year' in array
        'project_years' based on the associated 'year_inflation'. The result is then
        added by unity. This parameter sets up the second index to slice the 'inflation_rate_arr',
    (4) Slice 'inflation_rate_arr' according to 'id_start' and 'id_end'. Add the results
        by unity, then multiple the associated elements.
    (5) Cost adjustment is undertaken by multiplication: 'cost' * 'mult'.
    """

    # Prepare attribute year_inflation
    if year_inflation is None:
        year_inflation = np.repeat(start_year, len(cost))

    else:
        if not isinstance(year_inflation, np.ndarray):
            raise TaxInflationException(
                f"Attribute year_inflation must be given as a numpy.ndarray, "
                f"not as a/an {year_inflation.__class__.__qualname__}"
            )

        if len(year_inflation) != len(cost):
            raise TaxInflationException(
                f"Unequal length of arrays: "
                f"cost: {len(cost)}, "
                f"year_inflation: {len(year_inflation)}"
            )

    year_inflation = year_inflation.astype(np.int64)

    # Raise an error: year_inflation is before the start year of the project
    if np.min(year_inflation) < start_year:
        raise TaxInflationException(
            f"year_inflation ({np.min(year_inflation)}) is before the start year "
            f"of the project ({start_year})"
        )

    # Raise an error: year_inflation is after the end year of the project
    if np.max(year_inflation) > end_year:
        raise TaxInflationException(
            f"year_inflation ({np.max(year_inflation)}) is after the end year "
            f"of the project ({end_year})"
        )

    # Create an array of inflation_rate
    inflation_rate_arr = check_input(target_func=project_years, param=inflation_rate)

    # Specify the start and end indices to slice the inflation rate array
    id_start = (
        np.array(
            [
                np.argwhere(year_inflation[i] == project_years).ravel()
                for i, _ in enumerate(year_inflation)
            ]
        ).ravel() + 1
    ).astype(np.int64)

    id_end = (
        (expense_year - year_inflation) + (year_inflation - start_year) + 1
    ).astype(np.int64)

    # Multipliers to adjust cost by inflation
    mult = np.array(
        [
            np.prod(1.0 + inflation_rate_arr[id_start[i]:id_end[i]])
            for i, _ in enumerate(id_start)
        ]
    )

    return cost * mult


def apply_cost_adjustment(
    start_year: int,
    cost: np.ndarray,
    expense_year: np.ndarray,
    project_years: np.ndarray,
    tax_portion: np.ndarray,
    tax_rate: np.ndarray | float,
    tax_discount: float,
    inflation_rate: np.ndarray | float,
    year_ref: np.ndarray,
) -> np.ndarray:
    """
    Adjusts cost based on inflation and tax scheme over the specified time period.

    Parameters
    ----------
    start_year : int
        The start year of the project.
    cost : np.ndarray
        An array of costs to be adjusted.
    expense_year : np.ndarray
        An array specifying the expense years for each cost element.
    project_years : np.ndarray
        An array of project years.
    tax_portion: np.ndarray
        The portion of 'cost' that is subject to tax (VAT or LBT).
        Must be an array of length equals to the length of 'cost' array.
    tax_rate: np.ndarray | float
        The tax rate to apply (VAT or LBT). Can be a single value or an array.
    tax_discount: float
        The tax discount to apply (VAT or LBT).
    inflation_rate: np.ndarray | float
        The inflation rate to apply. Can be a single value or an array.
    year_ref : np.ndarray
        An array of reference year for inflation calculation.

    Returns
    -------
    cost_adjusted: np.ndarray
        An array of adjusted costs after applying inflation and tax scheme.

    Notes
    -----
    The core operations are as follows:
    (1) Apply cost adjustment due tax scheme by calling function get_cost_adjustment_by_tax().
    (2) Check 'inflation_rate' whether it is provided as an array or as a single value.
        If it is given as an array, the length of the array must be consistent with the
        duration of the project. If it is given as a single value, then create an array
        of length equal to project duration with all elements set equal to the single
        value.
    (3) Parameter 'id_start' identify the index location of 'year_ref' in array
        'project_years'. The result is then added by unity. This parameter sets up
        the first index to slice the 'inflation_rate_arr',
    (4) Parameter 'id_end' configure the index location of 'expense_year' in array
        'project_years' based on the associated 'year_ref'. The result is then added by unity.
        This parameter sets up the second index to slice the 'inflation_rate_arr',
    (5) Slice 'inflation_rate_arr' according to 'id_start' and 'id_end'. Add the results
        by unity, then multiple the associated elements.
    (6) Cost adjustment is undertaken by multiplication: 'cost_adjusted_by_tax' * 'mult'.
    """

    # Cost adjustment due to tax
    cost_adjusted = get_cost_adjustment_by_tax(
        start_year=start_year,
        cost=cost,
        expense_year=expense_year,
        project_years=project_years,
        tax_portion=tax_portion,
        tax_rate=tax_rate,
        tax_discount=tax_discount,
    )

    # Create array of inflation_rate
    inflation_rate_arr = check_input(target_func=project_years, param=inflation_rate)

    # Specify the start and end indices to slice the inflation_rate array
    id_start = np.array(
        [
            np.argwhere(year_ref[i] == project_years).ravel()[0] + 1
            for i, val in enumerate(year_ref)
        ]
    )

    id_end = (
        (expense_year - year_ref) + (year_ref - start_year) + 1
    ).astype("int")

    # Multipliers to adjust cost by inflation
    mult = np.array(
        [
            np.prod(1.0 + inflation_rate_arr[id_start[i]:id_end[i]])
            for i, val in enumerate(id_start)
        ]
    )

    return cost_adjusted * mult


def calc_indirect_tax(
    start_year: int,
    cost: np.ndarray,
    expense_year: np.ndarray,
    project_years: np.ndarray,
    tax_portion: np.ndarray,
    tax_rate: np.ndarray | float,
    tax_discount: np.ndarray,
) -> np.ndarray:
    """
    Calculate the indirect tax on project costs.

    This function calculates the indirect tax by applying the specified tax portion,
    tax rates, and tax discount to the project costs for the corresponding expense years.

    Parameters
    ----------
    start_year : int
        The start year of the project. Used to calculate the tax rate indices.
    cost : np.ndarray
        A NumPy array representing the project costs to which the indirect tax will be applied.
    expense_year : np.ndarray
        A NumPy array representing the years during which expenses occur. Used to
        match tax rates to the corresponding years.
    project_years: np.ndarray
        A NumPy array representing project years.
    tax_portion : np.ndarray
        A NumPy array representing the portion of the cost that is subject to tax.
    tax_rate : np.ndarray | float
        A NumPy array of tax rates for each year of the project.
    tax_discount : np.ndarray
        A discount factor applied to the tax, reducing the overall tax impact.

    Returns
    -------
    np.ndarray
        A NumPy array containing the indirect tax amounts for each cost, adjusted for
        the tax portion, tax rates, and tax discount.

    Notes
    -----
    -   The function computes the tax rate indices based on the difference between
        `expense_year` and `start_year`.
    -   The indirect tax is calculated by applying the tax rate to the `tax_portion`
        of the `cost`, adjusted by the `tax_discount`.
    -   The function calculates the indirect tax using the formula:
            indirect_tax = cost * (tax_portion * tax_rate * (1.0 - tax_discount))
    """

    # Prepare attribute tax rate
    if not isinstance(tax_rate, (np.ndarray, float)):
        raise TaxInflationException(
            f"Attribute tax rate must be given as a float or as a numpy.ndarray, "
            f"not as a/an {tax_rate.__class__.__qualname__}"
        )

    else:
        tax_rate_arr = check_input(target_func=project_years, param=tax_rate)

    # Specify tax rate id
    tax_rate_ids = (expense_year - start_year).astype(np.int64)

    # Calculate indirect tax
    return cost * (tax_portion * tax_rate_arr[tax_rate_ids] * (1.0 - tax_discount))


def calc_distributed_cost(
    cost: np.ndarray,
    expense_year: np.ndarray,
    final_year: np.ndarray,
    project_years: np.ndarray,
    project_duration: int,
) -> np.ndarray:
    """
    Distribute the cost evenly over the project duration based on the expense and final years.

    This function decomposes the total cost of each element across multiple years,
    from the expense year to the final year. It then distributes the decomposed values
    over the specified project years.

    Parameters
    ----------
    cost : np.ndarray
        The total cost for each element. Must be a 1D array of floats.
    expense_year : np.ndarray
        The year in which each cost element is incurred. Must be a 1D array of integers.
    final_year : np.ndarray
        The final year for each cost element to be distributed. Must be a 1D array of integers.
    project_years : np.ndarray
        The array of project years used to index the distribution. Must be a 1D array of integers.
    project_duration : int
        The total duration of the project in years.

    Returns
    -------
    np.ndarray
        A 2D array where each column represents the distributed cost of an element over
        the project duration, with rows representing years.

    Notes
    -----
    -   The total cost for each element is split evenly across the number of years from
        `expense_year` to `final_year`.
    -   The cost is distributed across the range of `project_years` starting at `expense_year`
        and ending at `final_year`.
    -   The function returns a zero-filled matrix for years before the expense year or after
        the final year for each element.
    """

    # Prepare attribute cost
    if not isinstance(cost, np.ndarray):
        raise TaxInflationException(
            f"Argument cost must be given as a numpy.ndarray, "
            f"not as a/an {cost.__class__.__qualname__}"
        )

    # Prepare attribute expense_year
    if not isinstance(expense_year, np.ndarray):
        raise TaxInflationException(
            f"Argument expense_year must be given as a numpy.ndarray, "
            f"not as a/an {expense_year.__class__.__qualname__}"
        )

    # Unequal length of arrays
    if len(cost) != len(expense_year):
        raise TaxInflationException(
            f"Unequal length of arrays: "
            f"Expected: {len(expense_year)}, "
            f"Given: {len(cost)}"
        )

    cost = cost.astype(np.float64)

    # Total number of years to split/decompose each cost elements
    years_to_split = (final_year - expense_year + 1).astype(int)

    # Decomposed values for each cost elements
    cost_split = cost / years_to_split

    # The start and end indices to distribute each cost elements
    id_start = np.array(
        [
            np.argwhere(expense_year[i] == project_years).ravel()
            for i, _ in enumerate(expense_year)
        ]
    ).ravel()

    id_end = (id_start + years_to_split).astype(int)

    # Distributed values for each cost elements
    distributed_cost = np.zeros([project_duration, len(expense_year)], dtype=np.float64)

    for i, _ in enumerate(expense_year):
        distributed_cost[id_start[i]:id_end[i], i] = cost_split[i]

    return distributed_cost



"""
A collection of unit testings for OPEX class.
"""

import pytest
import numpy as np
from pyscnomics.econ.selection import FluidType, TaxType
from pyscnomics.econ.costs import OPEX, OPEXException, Intangible


def test_opex_incorrect_year_input():
    """A unit testing for incorrect data input: start year is after end year"""

    with pytest.raises(OPEXException):

        OPEX(
            start_year=2033,
            end_year=2030,
            fixed_cost=np.array([100, 50]),
            expense_year=np.array([2028, 2029]),
            cost_allocation=[FluidType.OIL, FluidType.OIL],
        )


def test_opex_unequal_length_of_data_input():
    """A unit testing for incorrect data input: unequal length of data input"""

    with pytest.raises(OPEXException):

        OPEX(
            start_year=2023,
            end_year=2030,
            fixed_cost=np.array([100, 50, 50]),
            expense_year=np.array([2028, 2029]),
            cost_allocation=[FluidType.OIL, FluidType.OIL],
            prod_rate=np.array([100, 50]),
            cost_per_volume=np.array([0.1, 0.1]),
        )

    with pytest.raises(OPEXException):

        OPEX(
            start_year=2023,
            end_year=2030,
            fixed_cost=np.array([100, 50]),
            expense_year=np.array([2028, 2029]),
            cost_allocation=[FluidType.OIL, FluidType.OIL],
            prod_rate=np.array([100, 100]),
        )

    with pytest.raises(OPEXException):

        OPEX(
            start_year=2023,
            end_year=2030,
            fixed_cost=np.array([100, 50]),
            expense_year=np.array([2028, 2029]),
            cost_allocation=[FluidType.OIL, FluidType.OIL],
            cost_per_volume=np.array([0.1, 0.1]),
        )

    with pytest.raises(OPEXException):

        OPEX(
            start_year=2023,
            end_year=2030,
            fixed_cost=np.array([100, 50]),
            expense_year=np.array([2028, 2029]),
            cost_allocation=[FluidType.OIL, FluidType.OIL],
            prod_rate=np.array([100, 50]),
            cost_per_volume=np.array([0.1, 0.1, 0.1]),
        )


def test_opex_incorrect_expense_year_input():
    """A unit testing for incorrect data input: expense year is after the end year of the project"""

    with pytest.raises(OPEXException):

        OPEX(
            start_year=2023,
            end_year=2030,
            fixed_cost=np.array([100, 50]),
            expense_year=np.array([2028, 2033]),
            cost_allocation=[FluidType.OIL, FluidType.OIL],
        )


def test_opex_comparison_error():
    """A unit testing for misuse of OPEX: comparing different instances/objects"""

    mangga_opex = OPEX(
        start_year=2023,
        end_year=2030,
        fixed_cost=np.array([100, 50]),
        expense_year=np.array([2028, 2029]),
        cost_allocation=[FluidType.OIL, FluidType.OIL],
    )

    mangga_intangible = Intangible(
        start_year=2023,
        end_year=2030,
        cost=np.array([100, 50]),
        expense_year=np.array([2028, 2029]),
        cost_allocation=[FluidType.OIL, FluidType.OIL],
    )

    # Execute testing
    check_equality = (mangga_opex == mangga_intangible)

    assert check_equality is False


def test_opex_comparison():
    """A unit testing for comparison between OPEX instances"""

    mangga_opex = OPEX(
        start_year=2023,
        end_year=2030,
        fixed_cost=np.array([100, 50]),
        expense_year=np.array([2028, 2029]),
        cost_allocation=[FluidType.OIL, FluidType.OIL],
    )

    apel_opex = OPEX(
        start_year=2023,
        end_year=2030,
        fixed_cost=np.array([100, 50]),
        expense_year=np.array([2028, 2029]),
        cost_allocation=[FluidType.OIL, FluidType.OIL],
    )

    jeruk_opex = OPEX(
        start_year=2023,
        end_year=2030,
        fixed_cost=np.array([50, 25]),
        expense_year=np.array([2028, 2029]),
        cost_allocation=[FluidType.OIL, FluidType.OIL],
    )

    nanas_opex = OPEX(
        start_year=2023,
        end_year=2030,
        fixed_cost=np.array([100, 50]),
        expense_year=np.array([2028, 2029]),
        cost_allocation=[FluidType.OIL, FluidType.OIL],
        prod_rate=np.array([100, 100]),
        cost_per_volume=np.array([0.1, 0.1]),
    )

    # Execute comparison testing
    assert mangga_opex == apel_opex
    assert mangga_opex != jeruk_opex
    assert jeruk_opex < apel_opex
    assert mangga_opex <= nanas_opex
    assert nanas_opex > mangga_opex
    assert apel_opex >= jeruk_opex


def test_opex_arithmetics_incorrect():
    """A unit testing for misuse of arithmetic operations upon an instance of OPEX"""

    mangga_opex = OPEX(
        start_year=2023,
        end_year=2030,
        fixed_cost=np.array([100, 50]),
        expense_year=np.array([2028, 2029]),
        cost_allocation=[FluidType.OIL, FluidType.OIL],
    )

    mangga_intangible = Intangible(
        start_year=2023,
        end_year=2030,
        cost=np.array([100, 50]),
        expense_year=np.array([2028, 2029]),
        cost_allocation=[FluidType.OIL, FluidType.OIL],
    )

    with pytest.raises(OPEXException):
        assert mangga_opex + 100

    with pytest.raises(OPEXException):
        assert mangga_opex - 50

    with pytest.raises(OPEXException):
        assert mangga_opex * mangga_intangible

    with pytest.raises(OPEXException):
        assert mangga_opex / 0


def test_opex_arithmetics():
    """A unit testing for arithmetic operations upon an instance of OPEX"""

    # Expected results
    add1 = [110, 110, 170, 170, 60, 60, 0, 0]
    add2 = [60, 60, 0, 0, 0, 0, 210, 210]
    mul1 = [0, 0, 0, 0, 0, 0, 420, 420]
    div1 = [30, 30, 0, 0, 0, 0, 0, 0]

    # Calculated results
    mangga_opex = OPEX(
        start_year=2023,
        end_year=2026,
        fixed_cost=np.array([100, 100, 100, 100]),
        expense_year=np.array([2023, 2024, 2025, 2026]),
        cost_allocation=[FluidType.OIL, FluidType.OIL, FluidType.OIL, FluidType.OIL],
        prod_rate=np.array([100, 100, 100, 100]),
        cost_per_volume=np.array([0.1, 0.1, 0.1, 0.1]),
    )

    apel_opex = OPEX(
        start_year=2025,
        end_year=2030,
        fixed_cost=np.array([50, 50, 50, 50]),
        expense_year=np.array([2025, 2026, 2027, 2028]),
        cost_allocation=[FluidType.OIL, FluidType.OIL, FluidType.OIL, FluidType.OIL],
        prod_rate=np.array([100, 100, 100, 100]),
        cost_per_volume=np.array([0.1, 0.1, 0.1, 0.1]),
    )

    nanas_opex = OPEX(
        start_year=2023,
        end_year=2030,
        fixed_cost=np.array([200, 200]),
        expense_year=np.array([2030, 2029]),
        cost_allocation=[FluidType.OIL, FluidType.OIL],
        prod_rate=np.array([100, 100]),
        cost_per_volume=np.array([0.1, 0.1]),
    )

    jeruk_opex = OPEX(
        start_year=2023,
        end_year=2030,
        fixed_cost=np.array([50, 50]),
        expense_year=np.array([2024, 2023]),
        cost_allocation=[FluidType.OIL, FluidType.OIL],
        prod_rate=np.array([100, 100]),
        cost_per_volume=np.array([0.1, 0.1]),
    )

    calc_add1 = (mangga_opex + apel_opex).expenditures_post_tax()
    calc_add2 = (nanas_opex + jeruk_opex).expenditures_post_tax()
    calc_mul1 = (nanas_opex * 2).expenditures_post_tax()
    calc_div1 = (jeruk_opex / 2).expenditures_post_tax()

    # Execute testing
    np.testing.assert_allclose(add1, calc_add1)
    np.testing.assert_allclose(add2, calc_add2)
    np.testing.assert_allclose(mul1, calc_mul1)
    np.testing.assert_allclose(div1, calc_div1)


def test_opex_expenditures():
    """A unit testing for expenditures method in OPEX class"""

    # Expected results
    expense1 = [0, 0, 0, 0, 0, 100, 50, 0]
    expense2 = [0, 0, 0, 0, 0, 110, 60, 0]

    # Calculated results
    mangga_opex = OPEX(
        start_year=2023,
        end_year=2030,
        fixed_cost=np.array([100, 50]),
        expense_year=np.array([2028, 2029]),
        cost_allocation=[FluidType.OIL, FluidType.OIL],
    )

    apel_opex = OPEX(
        start_year=2023,
        end_year=2030,
        fixed_cost=np.array([100, 50]),
        expense_year=np.array([2028, 2029]),
        cost_allocation=[FluidType.OIL, FluidType.OIL],
        prod_rate=np.array([100, 100]),
        cost_per_volume=np.array([0.1, 0.1]),
    )

    calc_expense1 = mangga_opex.expenditures_post_tax()
    calc_expense2 = apel_opex.expenditures_post_tax()

    # Execute testing
    np.testing.assert_allclose(expense1, calc_expense1)
    np.testing.assert_allclose(expense2, calc_expense2)


def test_opex_expenditures_with_tax_and_inflation():
    """ A unit testing for expenditures method in OPEX class """

    # Expected results
    case1 = np.array([0, 105, 110.25, 115.7625, 121.550625, 0, 0, 0])
    case2 = np.array([0, 102, 103, 104, 105, 0, 0, 0])
    case3 = np.array([0, 107.18504, 113.69756, 119.46108, 124.31724, 0, 0, 0])

    # Calculated results
    opex_mangga = OPEX(
        start_year=2023,
        end_year=2030,
        fixed_cost=np.array([100, 100, 100, 50]),
        expense_year=np.array([2024, 2025, 2026, 2027]),
        tax_portion=np.array([1, 1, 1, 1]),
        cost_allocation=[FluidType.OIL, FluidType.OIL, FluidType.OIL, FluidType.OIL],
    )

    opex_jeruk = OPEX(
        start_year=2023,
        end_year=2030,
        fixed_cost=np.array([50]),
        tax_portion=np.array([1]),
        expense_year=np.array([2027]),
        cost_allocation=[FluidType.OIL],
    )

    opex_apel = OPEX(
        start_year=2023,
        end_year=2030,
        fixed_cost=np.array([100, 100, 100, 100]),
        expense_year=np.array([2024, 2025, 2026, 2027]),
        cost_allocation=[FluidType.OIL, FluidType.OIL, FluidType.OIL, FluidType.OIL],
        tax_portion=np.array([0.36, 0.36, 0.36, 0.36]),
        tax_discount=0.743,
    )

    opex_add = opex_mangga + opex_jeruk

    case1_calc = opex_add.expenditures_post_tax(inflation_rate=0.05)
    case2_calc = opex_add.expenditures_post_tax(
        tax_rate=np.array([0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.07, 0.08]),
    )
    case3_calc = opex_apel.expenditures_post_tax(
        tax_rate=np.array([0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.07, 0.08]),
        inflation_rate=np.array([0.08, 0.07, 0.06, 0.05, 0.04, 0.03, 0.02, 0.01]),
    )

    # Execute testing
    np.testing.assert_allclose(case1, case1_calc)
    np.testing.assert_allclose(case2, case2_calc)
    np.testing.assert_allclose(case3, case3_calc)



"""
A unit testing for revenue module.
"""

import numpy as np
import pytest

from pyscnomics.econ.revenue import LiftingException, Lifting
from pyscnomics.econ.costs import FluidType


def test_lifting_incorrect_data_input():
    """
    Initial check on inappropriate data input.

    The test should raise a ValueError when the project duration is
    less than the length of lifting data.
    """

    with pytest.raises(LiftingException):
        Lifting(
            start_year=2023,
            end_year=2027,
            lifting_rate=np.array([50, 50, 50, 50, 50, 50, 50, 50]),
            price=np.array([10, 10, 10, 10, 10, 10, 10, 10]),
            fluid_type=FluidType.OIL,
            prod_year=np.array([2023, 2024, 2025, 2026, 2027, 2028, 2029, 2030]),
        )

    with pytest.raises(LiftingException):
        Lifting(
            start_year=2023,
            end_year=2030,
            lifting_rate=np.array([50, 50, 50, 50, 50, 50, 50, 50]),
            price=np.array([10, 10, 10, 10, 10, 10, 10, 10]),
            fluid_type=FluidType.OIL,
            prod_year=np.array([2021, 2022, 2023, 2024, 2025, 2026, 2027, 2028]),
        )

    with pytest.raises(LiftingException):
        Lifting(
            start_year=2023,
            end_year=2030,
            lifting_rate=np.array([50, 50, 50, 50, 50, 50, 50, 50]),
            price=np.array([10, 10, 10, 10, 10, 10, 10, 10]),
            fluid_type=FluidType.OIL,
            prod_year=np.array([2023, 2024, 2025, 2026, 2027, 2028, 2029, 2031]),
        )


def test_lifting_comparison():
    """
    A unit test for comparison involving an instance of Lifting.
    """

    mangga_lifting = Lifting(
        start_year=2023,
        end_year=2030,
        lifting_rate=np.array([100, 100, 100]),
        price=np.array([10, 10, 10]),
        prod_year=np.array([2025, 2026, 2027]),
        fluid_type=FluidType.OIL,
        prod_rate=np.array([150, 150, 150]),
        prod_rate_baseline=np.array([50, 50, 50]),
    )

    jeruk_lifting = Lifting(
        start_year=2023,
        end_year=2030,
        lifting_rate=np.array([100, 100, 100]),
        price=np.array([10, 10, 10]),
        prod_year=np.array([2025, 2026, 2027]),
        fluid_type=FluidType.OIL,
        prod_rate=np.array([150, 150, 150]),
        prod_rate_baseline=np.array([50, 50, 50]),
    )

    apel_lifting = Lifting(
        start_year=2023,
        end_year=2030,
        lifting_rate=np.array([100, 100, 100]),
        price=np.array([10, 10, 10]),
        prod_year=np.array([2025, 2026, 2027]),
        fluid_type=FluidType.GAS,
        prod_rate=np.array([150, 150, 150]),
        prod_rate_baseline=np.array([50, 50, 50]),
    )

    nanas_lifting = Lifting(
        start_year=2023,
        end_year=2030,
        lifting_rate=np.array([50, 50, 50]),
        price=np.array([10, 10, 10]),
        prod_year=np.array([2025, 2026, 2027]),
        fluid_type=FluidType.OIL,
    )

    # Execute testing conditions
    assert mangga_lifting == jeruk_lifting
    assert mangga_lifting != apel_lifting
    assert nanas_lifting < mangga_lifting
    assert nanas_lifting <= jeruk_lifting
    assert mangga_lifting > nanas_lifting
    assert jeruk_lifting >= nanas_lifting


def test_lifting_arithmetics():
    """
    Test arithmetic operations involving an instance of Lifting.
    """

    # Specify the expected result
    add1 = np.array([0, 0, 0, 0, 1000, 1000, 1000, 0])
    add2 = np.array([500, 500, 0, 0, 1000, 1000, 1500, 500])

    mul1 = np.array([0, 0, 0, 0, 2000, 2000, 2000, 0])
    mul2 = np.array([0.0, 0.0, 0.0, 0.0, -1000.0, -1000.0, -1000.0, 0.0])

    div1 = np.array([250, 250, 0, 0, 0, 0, 250, 250])
    div2 = np.array([-50, -50, 0, 0, 0, 0, -50, -50])

    # Create an instance for OIL revenue
    mangga_lifting = Lifting(
        start_year=2023,
        end_year=2030,
        lifting_rate=np.array([100, 100, 100]),
        price=np.array([10, 10, 10]),
        prod_year=np.array([2027, 2028, 2029]),
        fluid_type=FluidType.OIL,
        prod_rate=np.array([200, 200, 200]),
    )

    nanas_lifting = Lifting(
        start_year=2023,
        end_year=2030,
        lifting_rate=np.array([50, 50, 50, 50]),
        price=np.array([10, 10, 10, 10]),
        prod_year=np.array([2029, 2030, 2024, 2023]),
        fluid_type=FluidType.OIL,
        prod_rate=np.array([200, 200, 200, 200]),
    )

    total_lifting = mangga_lifting + nanas_lifting

    calc_add1 = mangga_lifting.revenue()
    calc_add2 = total_lifting.revenue()
    calc_mul1 = mangga_lifting * 2
    calc_mul2 = -1 * mangga_lifting
    calc_div1 = nanas_lifting / 2
    calc_div2 = nanas_lifting / -10

    # Test whether the instance is equal to the expected result
    np.testing.assert_allclose(add1, calc_add1)
    np.testing.assert_allclose(add2, calc_add2)

    np.testing.assert_allclose(mul1, calc_mul1.revenue())
    np.testing.assert_allclose(mul2, calc_mul2.revenue())

    np.testing.assert_allclose(div1, calc_div1.revenue())
    np.testing.assert_allclose(div2, calc_div2.revenue())


def test_prepare_lifting_rate():
    """
    Test the preparation of lifting_rate attribute in the Lifting class.

    This function tests two scenarios:
    1.  Ensures that an inappropriate data input raises a `LiftingException`.
    2.  Verifies that the calculated lifting rate matches the expected lifting rate
        for correct input data.

    Raises
    ------
    LiftingException
        If the input data is inappropriate.

    Notes
    -----
    The function performs the following tests:
    1.  Attempts to create a `Lifting` instance with inappropriate data
        and checks if a `LiftingException` is raised.
    2.  Creates a `Lifting` instance with valid data and verifies that the calculated lifting rate
        matches the expected lifting rate using `np.testing.assert_allclose`.
    """
    # Raise an exception for an inappropriate data
    with pytest.raises(LiftingException):
        Lifting(
            start_year=2023,
            end_year=2030,
            lifting_rate=100,
            price=np.array([10, 10, 10]),
            prod_year=np.array([2024, 2025, 2026]),
            fluid_type=FluidType.OIL,
            ghv=np.array([1, 1, 1]),
            prod_rate=np.array([200, 200, 200]),
            prod_rate_baseline=np.array([50, 50, 50]),
        )

    # Expected result
    lifting_rate_mangga = np.array([100., 100., 100.])

    # Calculated result
    lifting_mangga = Lifting(
        start_year=2023,
        end_year=2030,
        lifting_rate=np.array([100, 100, 100]),
        price=np.array([10, 10, 10]),
        prod_year=np.array([2024, 2025, 2026]),
        fluid_type=FluidType.OIL,
        ghv=np.array([1, 1, 1]),
        prod_rate=np.array([200, 200, 200]),
        prod_rate_baseline=np.array([50, 50, 50]),
    )

    lifting_rate_mangga_calc = lifting_mangga.lifting_rate

    # Test whether expected == calculated
    np.testing.assert_allclose(lifting_rate_mangga, lifting_rate_mangga_calc)


def test_prepare_price():
    """
    Test the preparation of price attribute in the Lifting class.

    This function tests two scenarios:
    1. Ensures that an inappropriate data input raises a `LiftingException`.
    2. Verifies that the calculated price matches the expected price for correct input data.

    Raises
    ------
    LiftingException
        If the input data is inappropriate.

    Notes
    -----
    The function performs the following tests:
    1.  Attempts to create a `Lifting` instance with inappropriate data (non-array price)
        and checks if a `LiftingException` is raised.
    2.  Creates a `Lifting` instance with valid data and verifies that the calculated
        price matches the expected price using `np.testing.assert_allclose`.
    """
    # Raise an exception for an inappropriate data
    with pytest.raises(LiftingException):
        Lifting(
            start_year=2023,
            end_year=2030,
            lifting_rate=np.array([100, 100, 100]),
            price=10,
            prod_year=np.array([2024, 2025, 2026]),
            fluid_type=FluidType.GAS,
            ghv=np.array([1, 1, 1]),
            prod_rate=np.array([200, 200, 200]),
            prod_rate_baseline=np.array([50, 50, 50]),
        )

    # Expected result
    price_apel_expected = np.array([10., 10., 10.])

    # Calculated result
    lifting_apel = Lifting(
        start_year=2023,
        end_year=2030,
        lifting_rate=np.array([100, 100, 100]),
        price=np.array([10, 10, 10]),
        prod_year=np.array([2024, 2025, 2026]),
        fluid_type=FluidType.GAS,
        ghv=np.array([1, 1, 1]),
        prod_rate=np.array([200, 200, 200]),
        prod_rate_baseline=np.array([50, 50, 50]),
    )

    price_apel_calculated = lifting_apel.price

    # Test whether expected == calculated
    np.testing.assert_allclose(price_apel_expected, price_apel_calculated)


def test_prepare_prod_year():
    """
    Test the preparation of attribute prod_year in the Lifting class.

    This function tests two scenarios:
    1.  Ensures that an inappropriate data input raises a `LiftingException`.
    2.  Verifies that the calculated production year matches the expected production year
        for correct input data.

    Raises
    ------
    LiftingException
        If the input data is inappropriate.

    Notes
    -----
    The function performs the following tests:
    1.  Attempts to create a `Lifting` instance with inappropriate data (non-array production year)
        and checks if a `LiftingException` is raised.
    2.  Creates a `Lifting` instance with valid data and verifies that the calculated production year
        matches the expected production year using `np.testing.assert_allclose`.
    """
    # Raise an exception for an inappropriate data
    with pytest.raises(LiftingException):
        Lifting(
            start_year=2023,
            end_year=2030,
            lifting_rate=np.array([100, 100, 100]),
            price=np.array([10, 10, 10]),
            prod_year=[2024, 2025, 2026],
            fluid_type=FluidType.GAS,
            ghv=np.array([1, 1, 1]),
            prod_rate=np.array([200, 200, 200]),
            prod_rate_baseline=np.array([50, 50, 50]),
        )

    # Expected result
    prod_year_apel_expected = np.array([2024, 2025, 2026])

    # Calculated result
    lifting_apel = Lifting(
        start_year=2023,
        end_year=2030,
        lifting_rate=np.array([100, 100, 100]),
        price=np.array([10, 10, 10]),
        prod_year=np.array([2024, 2025, 2026]),
        fluid_type=FluidType.GAS,
        ghv=np.array([1, 1, 1]),
        prod_rate=np.array([200, 200, 200]),
        prod_rate_baseline=np.array([50, 50, 50]),
    )

    prod_year_apel_calculated = lifting_apel.prod_year

    # Test whether expected == calculated
    np.testing.assert_allclose(prod_year_apel_expected, prod_year_apel_calculated)


def test_prepare_ghv():
    """
    Test the preparation of attribute ghv in the Lifting class.

    This function tests two scenarios:
    1. Ensures that an inappropriate data input raises a `LiftingException`.
    2. Verifies that the calculated GHV matches the expected GHV for correct input data.

    Raises
    ------
    LiftingException
        If the input data is inappropriate.

    Notes
    -----
    The function performs the following tests:
    1.  Attempts to create a `Lifting` instance with inappropriate data (non-array GHV)
        and checks if a `LiftingException` is raised.
    2.  Creates two `Lifting` instances with valid data and verifies that the calculated GHV
        matches the expected GHV using `np.testing.assert_allclose`.
    """
    # Raise an exception for an inappropriate data
    with pytest.raises(LiftingException):
        Lifting(
            start_year=2023,
            end_year=2030,
            lifting_rate=np.array([100, 100, 100]),
            price=np.array([10, 10, 10]),
            prod_year=[2024, 2025, 2026],
            fluid_type=FluidType.GAS,
            ghv=1,
            prod_rate=np.array([200, 200, 200]),
            prod_rate_baseline=np.array([50, 50, 50]),
        )

    # Expected results
    ghv_expected1 = np.array([1., 1., 1.])
    ghv_expected2 = np.array([0.5, 0.5, 0.5])

    # Calculated results
    lifting_apel1 = Lifting(
        start_year=2023,
        end_year=2030,
        lifting_rate=np.array([100, 100, 100]),
        price=np.array([10, 10, 10]),
        prod_year=np.array([2024, 2025, 2026]),
        fluid_type=FluidType.GAS,
        prod_rate=np.array([200, 200, 200]),
        prod_rate_baseline=np.array([50, 50, 50]),
    )

    lifting_apel2 = Lifting(
        start_year=2023,
        end_year=2030,
        lifting_rate=np.array([100, 100, 100]),
        price=np.array([10, 10, 10]),
        prod_year=np.array([2024, 2025, 2026]),
        ghv=np.array([0.5, 0.5, 0.5]),
        fluid_type=FluidType.GAS,
        prod_rate=np.array([200, 200, 200]),
        prod_rate_baseline=np.array([50, 50, 50]),
    )

    ghv_calculated1 = lifting_apel1.ghv
    ghv_calculated2 = lifting_apel2.ghv

    # Test whether expected == calculated
    np.testing.assert_allclose(ghv_expected1, ghv_calculated1)
    np.testing.assert_allclose(ghv_expected2, ghv_calculated2)


def test_prepare_prod_rate():
    """
    Test the preparation of attribute prod_rate in the Lifting class.

    This function tests three scenarios:
    1.  Ensures that an inappropriate data input raises a `LiftingException`.
    2.  Verifies that the calculated production rate matches the expected production rate
        for correct input data.

    Raises
    ------
    LiftingException
        If the input data is inappropriate.

    Notes
    -----
    The function performs the following tests:
    1.  Attempts to create a `Lifting` instance with inappropriate data (non-array production rate)
        and checks if a `LiftingException` is raised.
    2.  Attempts to create a `Lifting` instance with mismatched production rate and
        baseline production rate arrays and checks if a `LiftingException` is raised.
    3.  Creates two `Lifting` instances with valid data and verifies that the calculated
        production rate matches the expected production rate using `np.testing.assert_allclose`.
    """
    # Raise an exception for an inappropriate data
    with pytest.raises(LiftingException):
        Lifting(
            start_year=2023,
            end_year=2030,
            lifting_rate=np.array([100, 100, 100]),
            price=np.array([10, 10, 10]),
            prod_year=[2024, 2025, 2026],
            fluid_type=FluidType.GAS,
            ghv=np.array([1, 1, 1]),
            prod_rate=200,
            prod_rate_baseline=np.array([50, 50, 50]),
        )

    # Raise an exception for an inappropriate data
    with pytest.raises(LiftingException):
        Lifting(
            start_year=2023,
            end_year=2030,
            lifting_rate=np.array([100, 100, 100]),
            price=np.array([10, 10, 10]),
            prod_year=[2024, 2025, 2026],
            fluid_type=FluidType.GAS,
            ghv=np.array([1, 1, 1]),
            prod_rate=np.array([50, 50, 50]),
            prod_rate_baseline=np.array([50, 50, 50]),
        )

    # Expected results
    prod_rate_expected1 = np.array([100., 100., 100.])
    prod_rate_expected2 = np.array([150., 150., 150.])

    # Calculated results
    lifting_apel1 = Lifting(
        start_year=2023,
        end_year=2030,
        lifting_rate=np.array([100, 100, 100]),
        price=np.array([10, 10, 10]),
        prod_year=np.array([2024, 2025, 2026]),
        fluid_type=FluidType.GAS,
        prod_rate_baseline=np.array([50, 50, 50]),
    )

    lifting_apel2 = Lifting(
        start_year=2023,
        end_year=2030,
        lifting_rate=np.array([100, 100, 100]),
        price=np.array([10, 10, 10]),
        prod_year=np.array([2024, 2025, 2026]),
        fluid_type=FluidType.GAS,
        prod_rate=np.array([150, 150, 150]),
        prod_rate_baseline=np.array([50, 50, 50]),
    )

    prod_rate_calculated1 = lifting_apel1.prod_rate
    prod_rate_calculated2 = lifting_apel2.prod_rate

    # Test whether expected == calculated
    np.testing.assert_allclose(prod_rate_expected1, prod_rate_calculated1)
    np.testing.assert_allclose(prod_rate_expected2, prod_rate_calculated2)


def test_prepare_prod_rate_baseline():
    """
    Test the preparation of attribute prod_rate_baseline in the Lifting class.

    This function tests two scenarios:
    1.  Ensures that an inappropriate data input raises a `LiftingException`.
    2.  Verifies that the calculated baseline production rate matches the expected baseline
        production rate for correct input data.

    Raises
    ------
    LiftingException
        If the input data is inappropriate.

    Notes
    -----
    The function performs the following tests:
    1.  Attempts to create a `Lifting` instance with inappropriate data
        (non-array baseline production rate) and checks if a `LiftingException` is raised.
    2.  Creates two `Lifting` instances with valid data and verifies that the calculated
        baseline production rate matches the expected baseline production rate using
        `np.testing.assert_allclose`.
    """
    # Raise an exception for an inappropriate data
    with pytest.raises(LiftingException):
        Lifting(
            start_year=2023,
            end_year=2030,
            lifting_rate=np.array([100, 100, 100]),
            price=np.array([10, 10, 10]),
            prod_year=[2024, 2025, 2026],
            fluid_type=FluidType.GAS,
            ghv=np.array([1, 1, 1]),
            prod_rate=np.array([150, 150, 150]),
            prod_rate_baseline=50,
        )

    # Expected results
    prod_rate_baseline_expected1 = np.array([0., 0., 0.])
    prod_rate_baseline_expected2 = np.array([50, 50, 50])

    # Calculated results
    lifting_apel1 = Lifting(
        start_year=2023,
        end_year=2030,
        lifting_rate=np.array([100, 100, 100]),
        price=np.array([10, 10, 10]),
        prod_year=np.array([2024, 2025, 2026]),
        ghv=np.array([1, 1, 1]),
        fluid_type=FluidType.GAS,
        prod_rate=np.array([150, 150, 150]),
    )

    lifting_apel2 = Lifting(
        start_year=2023,
        end_year=2030,
        lifting_rate=np.array([100, 100, 100]),
        price=np.array([10, 10, 10]),
        prod_year=np.array([2024, 2025, 2026]),
        ghv=np.array([1, 1, 1]),
        fluid_type=FluidType.GAS,
        prod_rate=np.array([150, 150, 150]),
        prod_rate_baseline=np.array([50, 50, 50]),
    )

    prod_rate_baseline_calculated1 = lifting_apel1.prod_rate_baseline
    prod_rate_baseline_calculated2 = lifting_apel2.prod_rate_baseline

    # Test whether expected == calculated
    np.testing.assert_allclose(prod_rate_baseline_expected1, prod_rate_baseline_calculated1)
    np.testing.assert_allclose(prod_rate_baseline_expected2, prod_rate_baseline_calculated2)


def test_prepare_prod_rate_total():
    """
    Test the preparation of attribute prod_rate_total in the Lifting class.

    This function verifies that the calculated total production rate matches the
    expected total production rate for given input data.

    Notes
    -----
    The function performs the following test:
    Creates a `Lifting` instance with valid data and verifies that the calculated total
    production rate matches the expected total production rate using `np.testing.assert_allclose`.
    """
    # Expected result
    prod_rate_total_expected = np.array([200., 200., 200.])

    # Calculated result
    lifting_apel = Lifting(
        start_year=2023,
        end_year=2030,
        lifting_rate=np.array([100, 100, 100]),
        price=np.array([10, 10, 10]),
        prod_year=np.array([2024, 2025, 2026]),
        ghv=np.array([1, 1, 1]),
        fluid_type=FluidType.GAS,
        prod_rate=np.array([150, 150, 150]),
        prod_rate_baseline=np.array([50, 50, 50]),
    )

    prod_rate_total_calculated = lifting_apel.prod_rate_total

    # Test whether expected == calculated
    np.testing.assert_allclose(prod_rate_total_expected, prod_rate_total_calculated)


def test_prepare_get_prod_rate_total_arr():
    """
    Test the `get_prod_rate_total_arr` method in the Lifting class.

    This function verifies that the calculated production rate total array matches
    the expected production rate total array for given input data.

    Notes
    -----
    The function performs the following test:
    Creates a `Lifting` instance with valid data and verifies that the calculated production rate
    total array matches the expected production rate total array using `np.testing.assert_allclose`.
    """
    # Expected result
    prod_rate_total_arr_expected = np.array([0., 200., 200., 0., 0., 0., 200., 200.])

    # Calculated result
    lifting_apel = Lifting(
        start_year=2023,
        end_year=2030,
        lifting_rate=np.array([100, 100, 100, 100]),
        price=np.array([10, 10, 10, 10]),
        prod_year=np.array([2024, 2025, 2029, 2030]),
        ghv=np.array([1, 1, 1, 1]),
        fluid_type=FluidType.GAS,
        prod_rate=np.array([150, 150, 150, 150]),
        prod_rate_baseline=np.array([50, 50, 50, 50]),
    )

    prod_rate_total_arr_calculated = lifting_apel.get_prod_rate_total_arr()

    # Test whether expected == calculated
    np.testing.assert_allclose(prod_rate_total_arr_expected, prod_rate_total_arr_calculated)



import numpy as np

from pyscnomics.dataset.sample import load_data
from pyscnomics.econ.selection import FTPTaxRegime, ContractSample
from pyscnomics.contracts.costrecovery import CostRecovery
from pyscnomics.contracts.grossplit import GrossSplit


class ObjectSampleException(Exception):
    """Exception to raise for a misuse of contract sample object generation"""

    pass


def generate_contract_sample(case: ContractSample) -> CostRecovery | GrossSplit:
    """
    Function to generate sample of contract object.

    Parameters
    ----------
    case: ContractSample
        The selection of the contract sample

    Returns
    -------
    psc: CostRecovery | GrossSplit
        The contract object

    """
    if case == ContractSample.CASE_1:
        # Defining the contract object
        psc = load_data(dataset_type='case1', contract_type='cost_recovery')

        # Defining the contract arguments
        tax_rate = 0.405
        ftp_tax_regime = FTPTaxRegime.DIRECT_MODE
        sunk_cost_reference_year = 2023
        post_uu_22_year2001 = True

        contract_arguments = {
            'effective_tax_rate': tax_rate,
            'ftp_tax_regime': ftp_tax_regime,
            'sunk_cost_reference_year': sunk_cost_reference_year,
            'post_uu_22_year2001': post_uu_22_year2001,
        }

        # Running the contract with contract arguments
        psc.run(**contract_arguments)

        return psc

    elif case == ContractSample.CASE_2:
        # Defining the contract object
        psc = load_data(dataset_type='case2', contract_type='cost_recovery')

        # Defining the contract arguments
        tax_rate = 0.405
        ftp_tax_regime = FTPTaxRegime.DIRECT_MODE
        sunk_cost_reference_year = 2023
        post_uu_22_year2001 = True

        contract_arguments = {
            'effective_tax_rate': tax_rate,
            'ftp_tax_regime': ftp_tax_regime,
            'sunk_cost_reference_year': sunk_cost_reference_year,
            'post_uu_22_year2001': post_uu_22_year2001,
        }

        # Running the contract with contract arguments
        psc.run(**contract_arguments)

        return psc

    elif case == ContractSample.CASE_3:
        # Defining the contract object
        psc = load_data(dataset_type='case3', contract_type='gross_split')

        # Defining the contract arguments
        tax_rate = 0.22
        sunk_cost_reference_year = 2022
        vat_rate = np.array(
            [0.11, 0.11, 0.11, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12,
             0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12, 0.12,
             0.12, 0.12])
        inflation_rate = 0.0
        inflation_rate_applied_to = None

        # Defining the contract arguments
        contract_arguments = {
            "effective_tax_rate": tax_rate,
            "sunk_cost_reference_year": sunk_cost_reference_year,
            "vat_rate": vat_rate,
            "inflation_rate": inflation_rate,
            "inflation_rate_applied_to": inflation_rate_applied_to,
        }

        # Running the contract with contract arguments
        psc.run(**contract_arguments)

        return psc
    else:
        raise ObjectSampleException(
            f"THE case argument {case} , "
            f"is should be one of: {[member.name for member in ContractSample]}"
        )

